"""
 Copyright (c) 2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import numpy as np
import torch
import torch.nn as nn

from nncf.layer_utils import COMPRESSION_MODULES
from nncf.utils import is_tracing_state, no_jit_trace


@COMPRESSION_MODULES.register()
class FilterPruningBlock(nn.Module):
    def __init__(self, size):
        super().__init__()
        self.register_buffer("_binary_filter_pruning_mask", torch.ones(size))

    @property
    def binary_filter_pruning_mask(self):
        return self._binary_filter_pruning_mask

    @binary_filter_pruning_mask.setter
    def binary_filter_pruning_mask(self, mask):
        with torch.no_grad():
            self._binary_filter_pruning_mask.set_(mask)

    def forward(self, conv_weight):
        if is_tracing_state():
            with no_jit_trace():
                return conv_weight
        return conv_weight


def broadcast_filter_mask(filter_mask, shape):
    broadcasted_shape = np.ones(len(shape), dtype=np.int64)
    broadcasted_shape[0] = filter_mask.size(0)
    broadcasted_filter_mask = torch.reshape(filter_mask, tuple(broadcasted_shape))
    return broadcasted_filter_mask


def inplace_apply_filter_binary_mask(filter_mask, conv_weight, module_name=""):
    """
    Inplace applying binary filter mask to weight (or bias) of the convolution
    (by first dim of the conv weight).
    :param filter_mask: binary mask (should have the same shape as first dim of conv weight)
    :param conv_weight: weight or bias of convolution
    :return: result with applied mask
    """
    if filter_mask.size(0) != conv_weight.size(0):
        raise RuntimeError("Shape of mask = {} for module {} isn't broadcastable to weight shape={}."
                           " ".format(filter_mask.shape, module_name, conv_weight.shape))
    broadcasted_filter_mask = broadcast_filter_mask(filter_mask, conv_weight.shape)
    return conv_weight.mul_(broadcasted_filter_mask)


def apply_filter_binary_mask(filter_mask, conv_weight, module_name=""):
    """
    Applying binary filter mask to weight (or bias) of the convolution (applying by first dim of the conv weight)
    without changing the weight.
    :param filter_mask: binary mask (should have the same shape as first dim of conv weight)
    :param conv_weight: weight or bias of convolution
    :return: result with applied mask
    """
    if filter_mask.size(0) != conv_weight.size(0):
        raise RuntimeError("Shape of mask = {} for module {} isn't broadcastable to weight shape={}."
                           " ".format(filter_mask.shape, module_name, conv_weight.shape))
    broadcasted_filter_mask = broadcast_filter_mask(filter_mask, conv_weight.shape)
    return broadcasted_filter_mask * conv_weight



import torch.nn as nn
from .registry import Registry


COMPRESSION_MODULES = Registry('compression modules')


class ProxyModule:
    def __init__(self, module):
        self._module = module

    def __getattr__(self, name):
        return getattr(self._module, name)


class _NNCFModuleMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _NNCFModuleMixin.add_mixin_fields(self)

    @staticmethod
    def add_mixin_fields(obj):
        obj.pre_ops = nn.ModuleDict()
        obj.post_ops = nn.ModuleDict()

    def get_pre_op(self, key):
        return self.pre_ops[key]

    def get_post_op(self, key):
        return self.post_ops[key]

    def register_pre_forward_operation(self, op):
        key = str(len(self.pre_ops))
        self.pre_ops[key] = op
        return key

    def remove_pre_forward_operation(self, key):
        return self.pre_ops.pop(key)

    def register_post_forward_operation(self, op):
        key = str(len(self.post_ops))
        self.post_ops[key] = op
        return key

    def remove_post_forward_operation(self, key):
        return self.post_ops.pop(key)

    def forward(self, *args):
        proxy_module = ProxyModule(self)
        for op in self.pre_ops.values():
            op_args = op(proxy_module, args)
            if op_args is not None:
                if not isinstance(op_args, tuple):
                    op_args = tuple([op_args])
                args = op_args
        results = super().forward.__func__(proxy_module, *args)
        for op in self.post_ops.values():
            op_results = op(proxy_module, results)
            if op_results is not None:
                results = op_results
        return results



"""
 Copyright (c) 2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
from copy import copy
from typing import List, Optional

import torch

from nncf.dynamic_graph.patch_pytorch import CustomTraceFunction, ForwardTraceOnly
from nncf.dynamic_graph.input_wrapping import MODEL_INPUT_OP_NAME
from nncf.dynamic_graph.version_agnostic_op_names import get_version_agnostic_name
from nncf.hw_config_op_names import HWConfigOpName
from nncf.registry import Registry


class OperatorMetatype:
    """Base class for grouping PyTorch operators based on their semantic meaning.
    Each derived class represents a single semantic group - for example, AddMetatype would
    group together '__iadd__', '__add__' and '__radd__' operations which all define elementwise
    tensor addition.
    Derived classes also specify which PyTorch functions in which modules should be patched
    and in what manner, so that the entire group of operations is visible in the internal graph
    representation. Grouping also allows efficient application of HW specifics to compression of
    certain operation groups.
    """
    name = ""

    # Wrapping specifications for operator calls of the following kind:
    # torch.nn.functional.conv2d
    torch_nn_functional_patch_spec = None  # type: Optional[PatchSpec]

    # Wrapping specifications for operator calls of the following kind:
    # torch.cat
    torch_module_patch_spec = None  # type: Optional[PatchSpec]

    # Wrapping specifications for operator calls of the following kind:
    # x = torch.Tensor(...)
    # x1 = x.view(...)
    torch_tensor_patch_spec = None  # type: Optional[PatchSpec]

    # Names of functions registered as operators via @register_operator to be associated
    # with this metatype
    external_op_names = []  # type: List[str]

    hw_config_names = []  # type: List[HWConfigOpName]

    subtypes = []  # type: List[OperatorSubtype]

    @classmethod
    def get_all_aliases(cls: 'OperatorMetatype') -> List[str]:
        # TODO: disambiguate overlapping function names
        retval = copy(cls.external_op_names)
        if cls.torch_nn_functional_patch_spec is not None:
            for fn_name in cls.torch_nn_functional_patch_spec.underlying_function_names:
                retval.append(fn_name)
        if cls.torch_module_patch_spec is not None:
            for fn_name in cls.torch_module_patch_spec.underlying_function_names:
                retval.append(fn_name)
        if cls.torch_tensor_patch_spec is not None:
            for fn_name in cls.torch_tensor_patch_spec.underlying_function_names:
                retval.append(fn_name)
        return retval

    @classmethod
    def determine_subtype(cls,
                          containing_module: Optional[torch.nn.Module] = None,
                          function_args=None,
                          functions_kwargs=None) -> Optional['OperatorSubtype']:
        matches = []
        for subtype in cls.subtypes:
            if subtype.matches(containing_module,
                               function_args,
                               functions_kwargs):
                matches.append(subtype)
        assert len(matches) <= 1, "Multiple subtypes match operator call " \
                                  "- cannot determine single subtype."
        if not matches:
            return None

        return matches[0]


class PatchSpec:
    def __init__(self,
                 underlying_function_names: List[str],
                 custom_trace_fn: CustomTraceFunction = None):
        """
        :param underlying_function_names: All function names in this list will be wrapped with NNCF
        wrappers that allow corresponding function calls to be registered in NNCF internal graph
        representation of the PyTorch model and to be afterwards considered for compression.
        :param custom_trace_fn: Will be called instead of the regular node search/insertion step
        during the corresponding operator call. Useful to specify this for nodes that have no effect on compression
        and therefore not vital to graph representation, but that should still be accounted for so that the
        graph representation does not become disjoint."""
        self.underlying_function_names = underlying_function_names
        self.custom_trace_fn = custom_trace_fn


class OperatorSubtype(OperatorMetatype):
    """Exact specialization of OperatorMetatype that can only be determined via operator argument
    inspection or owning module attribute inspection, and that may have specialized compression method
    configuration other than the one used for general operations having the type of OperatorMetatype."""

    @classmethod
    def matches(cls, containing_module: Optional[torch.nn.Module] = None,
                function_args=None,
                functions_kwargs=None) -> bool:
        raise NotImplementedError


class OperatorMetatypeRegistry(Registry):
    def __init__(self, name):
        super().__init__(name)
        self._op_name_to_op_meta_dict = {}

    def register(self, name=None):
        name_ = name
        super_register = super()._register

        def wrap(obj: 'OperatorMetatype'):
            cls_name = name_
            if cls_name is None:
                cls_name = obj.__name__
            super_register(obj, cls_name)
            op_names = obj.get_all_aliases()
            for name in op_names:
                name = get_version_agnostic_name(name)
                if name not in self._op_name_to_op_meta_dict:
                    self._op_name_to_op_meta_dict[name] = obj
                else:
                    assert self._op_name_to_op_meta_dict[name] == obj, \
                        "Inconsistent operator metatype registry - single patched op name maps to multiple metatypes!"
            return obj

        return wrap

    def get_operator_metatype_by_op_name(self, op_name: str) -> 'OperatorMetatype':
        if op_name not in self._op_name_to_op_meta_dict:
            return self._op_name_to_op_meta_dict["noop"]  # avoiding the dependency on the class def below
        return self._op_name_to_op_meta_dict[op_name]


OPERATOR_METATYPES = OperatorMetatypeRegistry("operator_metatypes")


@OPERATOR_METATYPES.register()
class NoopMetatype(OperatorMetatype):
    name = "noop"
    external_op_names = [name, MODEL_INPUT_OP_NAME]


@OPERATOR_METATYPES.register()
class DepthwiseConv1dSubtype(OperatorSubtype):
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]

    @classmethod
    def matches(cls, containing_module: Optional[torch.nn.Module] = None,
                function_args=None,
                functions_kwargs=None) -> bool:
        if containing_module.groups == containing_module.in_channels and containing_module.in_channels > 1:
            return True
        return False


@OPERATOR_METATYPES.register()
class Conv1dMetatype(OperatorMetatype):
    name = "conv1d"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    subtypes = [DepthwiseConv1dSubtype]
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class DepthwiseConv2dSubtype(OperatorSubtype):
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]

    @classmethod
    def matches(cls, containing_module: Optional[torch.nn.Module] = None,
                function_args=None,
                functions_kwargs=None) -> bool:

        if containing_module.groups == containing_module.in_channels and containing_module.in_channels > 1:
            return True
        return False


@OPERATOR_METATYPES.register()
class Conv2dMetatype(OperatorMetatype):
    name = "conv2d"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    torch_nn_functional_patch_spec = PatchSpec([name])
    subtypes = [DepthwiseConv2dSubtype]


@OPERATOR_METATYPES.register()
class DepthwiseConv3dSubtype(OperatorSubtype):
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]

    @classmethod
    def matches(cls, containing_module: Optional[torch.nn.Module] = None,
                function_args=None,
                functions_kwargs=None) -> bool:

        if containing_module.groups == containing_module.in_channels and containing_module.in_channels > 1:
            return True
        return False


@OPERATOR_METATYPES.register()
class Conv3dMetatype(OperatorMetatype):
    name = "conv3d"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    subtypes = [DepthwiseConv3dSubtype]
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class ConvTranspose2dMetatype(OperatorMetatype):
    name = "conv_transpose2d"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class ConvTranspose3dMetatype(OperatorMetatype):
    name = "conv_transpose3d"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class LinearMetatype(OperatorMetatype):
    name = "linear"
    hw_config_names = [HWConfigOpName.MATMUL]
    torch_nn_functional_patch_spec = PatchSpec([name])
    torch_module_patch_spec = PatchSpec(["addmm"])


@OPERATOR_METATYPES.register()
class HardTanhMetatype(OperatorMetatype):
    name = "hardtanh"
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class TanhMetatype(OperatorMetatype):
    name = "tanh"
    torch_nn_functional_patch_spec = PatchSpec([name])
    torch_module_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class ELUMetatype(OperatorMetatype):
    name = "elu"
    torch_nn_functional_patch_spec = PatchSpec([name, "elu_"])


@OPERATOR_METATYPES.register()
class PRELUMetatype(OperatorMetatype):
    name = "prelu"
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class LayerNormMetatype(OperatorMetatype):
    name = "layer_norm"
    torch_nn_functional_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.MVN]


@OPERATOR_METATYPES.register()
class GELUMetatype(OperatorMetatype):
    name = "gelu"
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class SigmoidMetatype(OperatorMetatype):
    name = "sigmoid"
    torch_nn_functional_patch_spec = PatchSpec([name])
    torch_module_patch_spec = PatchSpec([name])
    torch_tensor_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class AddMetatype(OperatorMetatype):
    name = "add"
    torch_tensor_patch_spec = PatchSpec([name,
                                         "__add__",
                                         "__iadd__",
                                         "__radd__"])
    torch_module_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.ADD]


@OPERATOR_METATYPES.register()
class SubMetatype(OperatorMetatype):
    name = "sub"
    torch_tensor_patch_spec = PatchSpec(["__sub__",
                                         "__isub__",
                                         "__rsub__"])
    hw_config_names = [HWConfigOpName.SUBTRACT]


@OPERATOR_METATYPES.register()
class MulMetatype(OperatorMetatype):
    name = "mul"
    torch_tensor_patch_spec = PatchSpec(["mul",
                                         "__mul__",
                                         "__imul__",
                                         "__rmul__"])
    hw_config_names = [HWConfigOpName.MULTIPLY]


@OPERATOR_METATYPES.register()
class DivMetatype(OperatorMetatype):
    name = "div"
    torch_module_patch_spec = PatchSpec([name])
    torch_tensor_patch_spec = PatchSpec(["__div__",
                                         "__idiv__",
                                         "__truediv__"])
    hw_config_names = [HWConfigOpName.DIVIDE]


@OPERATOR_METATYPES.register()
class ExpMetatype(OperatorMetatype):
    name = "exp"
    torch_module_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class ErfMetatype(OperatorMetatype):
    name = "erf"
    torch_module_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class MatMulMetatype(OperatorMetatype):
    name = "matmul"
    torch_module_patch_spec = PatchSpec([name, "bmm"])
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.MATMUL]



@OPERATOR_METATYPES.register()
class MeanMetatype(OperatorMetatype):
    name = "mean"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.REDUCEMEAN]


@OPERATOR_METATYPES.register()
class RoundMetatype(OperatorMetatype):
    name = "round"
    torch_tensor_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class DropoutMetatype(OperatorMetatype):
    name = "dropout"
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class ThresholdMetatype(OperatorMetatype):
    name = "threshold"
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class BatchNormMetatype(OperatorMetatype):
    name = "batch_norm"
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class AvgPool2dMetatype(OperatorMetatype):
    name = "avg_pool2d"
    hw_config_names = [HWConfigOpName.AVGPOOL]
    torch_nn_functional_patch_spec = PatchSpec([name, "adaptive_avg_pool2d"])


@OPERATOR_METATYPES.register()
class AvgPool3dMetatype(OperatorMetatype):
    name = "avg_pool3d"
    hw_config_names = [HWConfigOpName.AVGPOOL]
    torch_nn_functional_patch_spec = PatchSpec([name, "adaptive_avg_pool3d"])


@OPERATOR_METATYPES.register()
class MaxPool2dMetatype(OperatorMetatype):
    name = "max_pool2d"
    torch_nn_functional_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.MAXPOOL]


@OPERATOR_METATYPES.register()
class MaxPool3dMetatype(OperatorMetatype):
    name = "max_pool3d"
    torch_nn_functional_patch_spec = PatchSpec([name, "adaptive_max_pool3d"])
    hw_config_names = [HWConfigOpName.MAXPOOL]


@OPERATOR_METATYPES.register()
class MaxUnpool3dMetatype(OperatorMetatype):
    name = "max_unpool3d"
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class PadMetatype(OperatorMetatype):
    name = "pad"
    torch_nn_functional_patch_spec = PatchSpec([name], ForwardTraceOnly())


@OPERATOR_METATYPES.register()
class CatMetatype(OperatorMetatype):
    name = "cat"
    torch_module_patch_spec = PatchSpec([name, "stack"])
    hw_config_names = [HWConfigOpName.CONCAT]


@OPERATOR_METATYPES.register()
class RELUMetatype(OperatorMetatype):
    name = "relu"
    torch_module_patch_spec = PatchSpec([name, "relu_"])


@OPERATOR_METATYPES.register()
class MaxMetatype(OperatorMetatype):
    name = "max"
    torch_module_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.MAXIMUM,
                       HWConfigOpName.REDUCEMAX]


@OPERATOR_METATYPES.register()
class MinMetatype(OperatorMetatype):
    name = "min"
    torch_module_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.MINIMUM]


@OPERATOR_METATYPES.register()
class ARangeMetatype(OperatorMetatype):
    name = "arange"
    torch_module_patch_spec = PatchSpec([name], ForwardTraceOnly())


@OPERATOR_METATYPES.register()
class TransposeMetatype(OperatorMetatype):
    name = "transpose"
    torch_module_patch_spec = PatchSpec([name], ForwardTraceOnly())
    torch_tensor_patch_spec = PatchSpec([name, "permute"], ForwardTraceOnly())
    hw_config_names = [HWConfigOpName.TRANSPOSE]


@OPERATOR_METATYPES.register()
class GatherMetatype(OperatorMetatype):
    name = "gather"
    torch_module_patch_spec = PatchSpec(["index_select", "where"], ForwardTraceOnly())
    torch_tensor_patch_spec = PatchSpec(["index_select", "__getitem__"], ForwardTraceOnly())


@OPERATOR_METATYPES.register()
class ScatterMetatype(OperatorMetatype):
    name = "scatter"
    torch_tensor_patch_spec = PatchSpec(["masked_fill", "masked_fill_"])


@OPERATOR_METATYPES.register()
class ReshapeMetatype(OperatorMetatype):
    name = "reshape"
    torch_module_patch_spec = PatchSpec(["squeeze", "flatten", "unsqueeze"], ForwardTraceOnly())
    torch_tensor_patch_spec = PatchSpec([name, "view", "flatten", "squeeze", "unsqueeze"], ForwardTraceOnly())
    hw_config_names = [HWConfigOpName.RESHAPE,
                       HWConfigOpName.SQUEEZE,
                       HWConfigOpName.UNSQUEEZE,
                       HWConfigOpName.FLATTEN]


@OPERATOR_METATYPES.register()
class ContiguousMetatype(OperatorMetatype):
    name = "contiguous"
    torch_tensor_patch_spec = PatchSpec([name], ForwardTraceOnly())


@OPERATOR_METATYPES.register()
class SplitMetatype(OperatorMetatype):
    name = "split"
    torch_tensor_patch_spec = PatchSpec([name, "chunk"], ForwardTraceOnly())
    hw_config_names = [HWConfigOpName.SPLIT]


@OPERATOR_METATYPES.register()
class ExpandMetatype(OperatorMetatype):
    name = "expand"
    torch_tensor_patch_spec = PatchSpec([name], ForwardTraceOnly())


# Non-quantizable ops
@OPERATOR_METATYPES.register()
class EmbeddingMetatype(OperatorMetatype):
    name = "embedding"
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class SoftmaxMetatype(OperatorMetatype):
    name = "softmax"
    torch_nn_functional_patch_spec = PatchSpec([name])


@OPERATOR_METATYPES.register()
class LessMetatype(OperatorMetatype):
    name = "__lt__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.LESS]


@OPERATOR_METATYPES.register()
class LessEqualMetatype(OperatorMetatype):
    name = "__le__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.LESSEQUAL]


@OPERATOR_METATYPES.register()
class GreaterMetatype(OperatorMetatype):
    name = "__gt__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.GREATER]


@OPERATOR_METATYPES.register()
class GreaterEqualMetatype(OperatorMetatype):
    name = "__ge__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.GREATEREQUAL]


@OPERATOR_METATYPES.register()
class ModMetatype(OperatorMetatype):
    name = "__mod__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.FLOORMOD]


@OPERATOR_METATYPES.register()
class EqualsMetatype(OperatorMetatype):
    name = "__eq__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.EQUAL]


@OPERATOR_METATYPES.register()
class NotEqualMetatype(OperatorMetatype):
    name = "__ne__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.NOTEQUAL]


@OPERATOR_METATYPES.register()
class LogicalOrMetatype(OperatorMetatype):
    name = "__or__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.LOGICALOR]


@OPERATOR_METATYPES.register()
class LogicalXorMetatype(OperatorMetatype):
    name = "__xor__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.LOGICALXOR]


@OPERATOR_METATYPES.register()
class LogicalAndMetatype(OperatorMetatype):
    name = "__and__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.LOGICALAND]


@OPERATOR_METATYPES.register()
class LogicalNotMetatype(OperatorMetatype):
    name = "logical_not_"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.LOGICALNOT]


@OPERATOR_METATYPES.register()
class PowerMetatype(OperatorMetatype):
    name = "__pow__"
    torch_tensor_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.POWER]


@OPERATOR_METATYPES.register()
class InterpolateMetatype(OperatorMetatype):
    name = "interpolate"
    torch_nn_functional_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.INTERPOLATE]


@OPERATOR_METATYPES.register()
class RepeatMetatype(OperatorMetatype):
    name = "repeat_interleave"
    torch_module_patch_spec = PatchSpec([name])
    hw_config_names = [HWConfigOpName.TILE]


@OPERATOR_METATYPES.register()
class CloneMetatype(OperatorMetatype):
    name = "clone"
    torch_tensor_patch_spec = PatchSpec([name], ForwardTraceOnly())

@OPERATOR_METATYPES.register()
class PixelShuffleMetatype(OperatorMetatype):
    name = "pixel_shuffle"
    torch_nn_functional_patch_spec = PatchSpec([name])



import queue
from typing import List

import numpy as np
import torch

from nncf.dynamic_graph.context import no_nncf_trace
from nncf.nncf_logger import logger as nncf_logger
from .layers import BaseQuantizer
from ..utils import get_flat_tensor_contents_string


class QuantizeRangeInitializer:
    def __init__(self, quantize_module: BaseQuantizer, num_init_steps: int):
        self.quantize_module = quantize_module
        self.device = next(self.quantize_module.parameters()).device
        self.scale_shape = self.quantize_module.scale_shape
        self.num_init_steps = num_init_steps
        self.num_register_init_steps = 0

    def register_input(self, x: torch.Tensor):
        raise NotImplementedError

    def reset(self):
        raise NotImplementedError

    def forward_hook(self, module, input_, output):
        if self.num_register_init_steps < self.num_init_steps:
            self.register_input(input_[0])
            self.num_register_init_steps += 1

    def apply_init(self):
        raise NotImplementedError


def max_reduce_like(input_, ref_tensor_shape):
    numel = np.prod(ref_tensor_shape)
    if numel == 1:
        return input_.max()
    tmp_max = input_
    for dim_idx, dim in enumerate(ref_tensor_shape):
        if dim == 1:
            tmp_max, _ = torch.max(tmp_max, dim_idx, keepdim=True)
    return tmp_max


def min_reduce_like(input_, ref_tensor_shape):
    numel = np.prod(ref_tensor_shape)
    if numel == 1:
        return input_.min()
    tmp_min = input_
    for dim_idx, dim in enumerate(ref_tensor_shape):
        if dim == 1:
            tmp_min, _ = torch.min(tmp_min, dim_idx, keepdim=True)
    return tmp_min


def get_channel_count_and_dim_idx(scale_shape: List[int]):
    channel_dim_idx = 0
    channel_count = 1
    for dim_idx, dim in enumerate(scale_shape):
        if dim != 1:
            channel_dim_idx = dim_idx
            channel_count = dim
    return channel_count, channel_dim_idx


def expand_like(input_: torch.Tensor, scale_shape: List[int]):
    retval = input_
    count, idx = get_channel_count_and_dim_idx(scale_shape)
    assert input_.numel() == count
    assert len(input_.size()) == 1
    for _ in range(0, idx):
        retval = retval.unsqueeze(0)
    for _ in range(idx + 1, len(scale_shape)):
        retval = retval.unsqueeze(-1)
    return retval


def split_into_channels(input_: np.ndarray, scale_shape: List[int]) -> List[np.ndarray]:
    channel_count, channel_dim_idx = get_channel_count_and_dim_idx(scale_shape)
    channel_first_tensor = np.moveaxis(input_, channel_dim_idx, 0)
    if channel_count == 1:
        return [channel_first_tensor]

    ret_list = []
    for i in range(channel_count):
        ret_list.append(channel_first_tensor[i, ...])
    return ret_list


class MinMaxInitializer(QuantizeRangeInitializer):
    def __init__(self, quantize_module: 'BaseQuantizer', num_init_steps: int, log_module_name: str = None):
        super().__init__(quantize_module, num_init_steps)
        self.min_values = torch.ones(self.scale_shape).to(self.device) * np.inf
        self.max_values = torch.ones(self.scale_shape).to(self.device) * (-np.inf)
        self.log_module_name = log_module_name

    def register_input(self, x: torch.Tensor):
        with no_nncf_trace():
            self.min_values = torch.min(min_reduce_like(x, self.scale_shape),
                                        self.min_values)
            self.max_values = torch.max(max_reduce_like(x, self.scale_shape),
                                        self.max_values)

    def reset(self):
        self.min_values = torch.ones(self.scale_shape).to(self.device) * np.inf
        self.max_values = torch.ones(self.scale_shape).to(self.device) * (-np.inf)

    def apply_init(self):
        nncf_logger.debug("Statistics: min={} max={}".format(get_flat_tensor_contents_string(self.min_values),
                                                             get_flat_tensor_contents_string(self.max_values)))
        self.quantize_module.apply_minmax_init(self.min_values, self.max_values, self.log_module_name)


class MeanMinMaxInitializer(QuantizeRangeInitializer):
    def __init__(self, quantize_module: 'BaseQuantizer', num_init_steps: int, log_module_name: str = None):
        super().__init__(quantize_module, num_init_steps)
        self.log_module_name = log_module_name
        self.all_min_values = []
        self.all_max_values = []

    def register_input(self, x: torch.Tensor):
        with no_nncf_trace():
            self.all_min_values.append(min_reduce_like(x, self.scale_shape))
            self.all_max_values.append(max_reduce_like(x, self.scale_shape))

    def reset(self):
        self.all_min_values.clear()
        self.all_max_values.clear()

    def apply_init(self):
        min_values = torch.ones(self.scale_shape).to(self.device) * np.inf
        max_values = torch.ones(self.scale_shape).to(self.device) * (-np.inf)
        if self.all_min_values:
            stacked_min = torch.stack(self.all_min_values)
            min_values = stacked_min.mean(dim=0).view(self.scale_shape)
        if self.all_max_values:
            stacked_max = torch.stack(self.all_max_values)
            max_values = stacked_max.mean(dim=0).view(self.scale_shape)
        nncf_logger.debug("Statistics: min={} max={}".format(get_flat_tensor_contents_string(min_values),
                                                             get_flat_tensor_contents_string(max_values)))
        self.quantize_module.apply_minmax_init(min_values, max_values, self.log_module_name)


def get_per_channel_history(raw_input_history: queue.Queue, scale_shape: List[int], discard_zeros=False) -> List:
    channel_count, _ = get_channel_count_and_dim_idx(scale_shape)
    per_channel_history = [None for i in range(channel_count)]
    while not raw_input_history.empty():
        entry = raw_input_history.get()
        split = split_into_channels(entry, scale_shape)
        for i in range(channel_count):
            flat_channel_split = split[i].flatten()

            if discard_zeros:
                # For post-RELU quantizers exact zeros may prevail and lead to
                # zero mean and MAD - discard them
                flat_channel_split = flat_channel_split[flat_channel_split != 0]

            if per_channel_history[i] is None:
                per_channel_history[i] = flat_channel_split
            else:
                per_channel_history[i] = np.concatenate([per_channel_history[i], flat_channel_split])
    return per_channel_history


class ThreeSigmaInitializer(QuantizeRangeInitializer):
    def __init__(self, quantize_module: 'BaseQuantizer', num_init_steps: int, log_module_name: str = None):
        super().__init__(quantize_module, num_init_steps)
        self.input_history = queue.Queue()
        self.log_module_name = log_module_name

    def register_input(self, x: torch.Tensor):
        with no_nncf_trace():
            self.input_history.put(x.detach().cpu().numpy())

    def reset(self):
        self.input_history = queue.Queue()

    def apply_init(self):
        self.medians = torch.ones(self.scale_shape).to(self.device)
        self.median_absolute_deviations = torch.ones(self.scale_shape).to(self.device)

        per_channel_history = get_per_channel_history(self.input_history, self.scale_shape,
                                                      discard_zeros=True)
        per_channel_median = [np.median(channel_hist) for channel_hist in per_channel_history]
        per_channel_mad = []
        for idx, median in enumerate(per_channel_median):
            # Constant factor depends on the distribution form - assuming normal
            per_channel_mad.append(1.4826 * np.median(abs(per_channel_history[idx] - median)))

        numpy_median = np.asarray(per_channel_median)
        numpy_mad = np.asarray(per_channel_mad)
        median_tensor = torch.from_numpy(numpy_median).to(self.device, dtype=torch.float)
        mad_tensor = torch.from_numpy(numpy_mad).to(self.device, dtype=torch.float)

        median_tensor = expand_like(median_tensor, self.scale_shape)
        mad_tensor = expand_like(mad_tensor, self.scale_shape)

        nncf_logger.debug("Statistics: median={} MAD={}".format(get_flat_tensor_contents_string(median_tensor),
                                                                get_flat_tensor_contents_string(mad_tensor)))
        self.quantize_module.apply_minmax_init(median_tensor - 3 * mad_tensor, median_tensor + 3 * mad_tensor,
                                               self.log_module_name)


class PercentileInitializer(QuantizeRangeInitializer):
    def __init__(self, quantize_module: 'BaseQuantizer',
                 num_init_steps: int,
                 min_percentile: float,
                 max_percentile: float,
                 log_module_name: str = None):
        super().__init__(quantize_module, num_init_steps)
        self.input_history = queue.Queue()
        self.log_module_name = log_module_name
        self.min_percentile = min_percentile  # NB: Both min_percentile and max_percentile
        self.max_percentile = max_percentile  # are valued between 0 and 100

    def register_input(self, x: torch.Tensor):
        with no_nncf_trace():
            self.input_history.put(x.detach().cpu().numpy())

    def reset(self):
        self.input_history = queue.Queue()

    def apply_init(self):
        self.min_values = torch.ones(self.scale_shape).to(self.device) * np.inf
        self.max_values = torch.ones(self.scale_shape).to(self.device) * (-np.inf)

        per_channel_history = get_per_channel_history(self.input_history, self.scale_shape)
        per_channel_min_percentiles = [np.percentile(channel_hist, self.min_percentile) for channel_hist in
                                       per_channel_history]
        per_channel_max_percentiles = [np.percentile(channel_hist, self.max_percentile) for channel_hist in
                                       per_channel_history]

        numpy_mins = np.asarray(per_channel_min_percentiles)
        numpy_maxs = np.asarray(per_channel_max_percentiles)
        mins_tensor = torch.from_numpy(numpy_mins).to(self.device, dtype=torch.float)
        maxs_tensor = torch.from_numpy(numpy_maxs).to(self.device, dtype=torch.float)

        mins_tensor = expand_like(mins_tensor, self.scale_shape)
        maxs_tensor = expand_like(maxs_tensor, self.scale_shape)

        nncf_logger.debug("Statistics: Min ({}%th) percentile = {},"
                          " Max ({}%th) percentile = {}".format(self.min_percentile,
                                                                get_flat_tensor_contents_string(mins_tensor),
                                                                self.max_percentile,
                                                                get_flat_tensor_contents_string(maxs_tensor)))
        self.quantize_module.apply_minmax_init(mins_tensor,
                                               maxs_tensor, self.log_module_name)



"""
 Copyright (c) 2019-2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from typing import List

import torch

from nncf.algo_selector import COMPRESSION_ALGORITHMS
from nncf.compression_method_api import CompressionAlgorithmController, CompressionLevel
from nncf.nncf_network import NNCFNetwork
from nncf.sparsity.base_algo import BaseSparsityAlgoBuilder, BaseSparsityAlgoController, SparseModuleInfo
from nncf.sparsity.layers import BinaryMask
from nncf.sparsity.magnitude.functions import WEIGHT_IMPORTANCE_FUNCTIONS, calc_magnitude_binary_mask
from nncf.sparsity.schedulers import SPARSITY_SCHEDULERS


@COMPRESSION_ALGORITHMS.register('magnitude_sparsity')
class MagnitudeSparsityBuilder(BaseSparsityAlgoBuilder):
    def create_weight_sparsifying_operation(self, module):
        return BinaryMask(module.weight.size())

    def build_controller(self, target_model: NNCFNetwork) -> CompressionAlgorithmController:
        params = self.config.get("params", {})
        return MagnitudeSparsityController(target_model, self._sparsified_module_info,
                                           self.config,
                                           params.get('weight_importance', 'normed_abs'))


class MagnitudeSparsityController(BaseSparsityAlgoController):
    def __init__(self, target_model: NNCFNetwork,
                 sparsified_module_info: List[SparseModuleInfo],
                 config, weight_importance: str):
        super().__init__(target_model, sparsified_module_info)
        self.config = config
        params = self.config.get("params", {})
        self.weight_importance = WEIGHT_IMPORTANCE_FUNCTIONS.get(weight_importance)
        sparsity_level_mode = params.get("sparsity_level_setting_mode", "global")
        self._scheduler = None
        if sparsity_level_mode == 'global':
            scheduler_cls = SPARSITY_SCHEDULERS.get(params.get("schedule", "polynomial"))
            self._scheduler = scheduler_cls(self, params)

    def statistics(self):
        stats = super().statistics()
        stats['sparsity_threshold'] =\
             self._select_threshold(self.sparsity_rate_for_sparsified_modules, self.sparsified_module_info)
        return stats

    def freeze(self):
        for layer in self.sparsified_module_info:
            layer.operand.frozen = True

    def set_sparsity_level(self, sparsity_level, target_sparsified_module_info: SparseModuleInfo = None):
        if sparsity_level >= 1 or sparsity_level < 0:
            raise AttributeError(
                'Sparsity level should be within interval [0,1), actual value to set is: {}'.format(sparsity_level))
        if target_sparsified_module_info is None:
            target_sparsified_module_info_list = self.sparsified_module_info # List[SparseModuleInfo]
        else:
            target_sparsified_module_info_list = [target_sparsified_module_info]
        threshold = self._select_threshold(sparsity_level, target_sparsified_module_info_list)
        self._set_masks_for_threshold(threshold, target_sparsified_module_info_list)
        self.run_batchnorm_adaptation(self.config)

    def _select_threshold(self, sparsity_level, target_sparsified_module_info_list):
        all_weights = self._collect_all_weights(target_sparsified_module_info_list)
        if not all_weights:
            return 0.0
        all_weights_tensor, _ = torch.cat(all_weights).sort()
        threshold = all_weights_tensor[int((all_weights_tensor.size(0) - 1) * sparsity_level)].item()
        return threshold

    def _set_masks_for_threshold(self, threshold_val, target_sparsified_module_info_list):
        for layer in target_sparsified_module_info_list:
            if not layer.operand.frozen:
                layer.operand.binary_mask = calc_magnitude_binary_mask(layer.module.weight,
                                                                       self.weight_importance,
                                                                       threshold_val)


    def _collect_all_weights(self, target_sparsified_module_info_list: List[SparseModuleInfo]):
        all_weights = []
        for minfo in target_sparsified_module_info_list:
            all_weights.append(self.weight_importance(minfo.module.weight).view(-1))
        return all_weights

    def create_weight_sparsifying_operation(self, module):
        return BinaryMask(module.weight.size())

    def compression_level(self) -> CompressionLevel:
        if self.scheduler is not None:
            return self.scheduler.compression_level()
        return CompressionLevel.NONE



"""
 Copyright (c) 2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
from copy import deepcopy
from os import path as osp
from typing import Callable, Any, Tuple, List, Dict

from nncf.checkpoint_loading import load_state
from nncf.hw_config import HWConfigType
from torch.nn import Module

from nncf.compression_method_api import CompressionAlgorithmController, CompressionAlgorithmBuilder
from nncf.config import NNCFConfig
from nncf.debug import is_debug, set_debug_log_dir
from nncf.dynamic_graph.graph_builder import GraphBuilder, create_input_infos, create_dummy_forward_fn
from nncf.nncf_network import NNCFNetwork
from nncf.utils import is_main_process
from nncf.algo_selector import COMPRESSION_ALGORITHMS
from nncf.quantization.algo import QuantizerSetupType
from nncf.hw_config import HW_CONFIG_TYPE_TARGET_DEVICE_MAP

from nncf.nncf_logger import logger


def get_compression_algorithm(config):
    algorithm_key = config.get('algorithm', 'NoCompressionAlgorithmBuilder')
    logger.info("Creating compression algorithm: {}".format(algorithm_key))
    return COMPRESSION_ALGORITHMS.get(algorithm_key)


def create_compression_algorithm_builders(config: NNCFConfig,
                                          should_init: bool = True) -> List[CompressionAlgorithmBuilder]:
    compression_config_json_section = config.get('compression', {})
    compression_config_json_section = deepcopy(compression_config_json_section)

    hw_config_type = None
    quantizer_setup_type_str = config.get("quantizer_setup_type", "propagation_based")
    quantizer_setup_type = QuantizerSetupType.from_str(quantizer_setup_type_str)
    if quantizer_setup_type == QuantizerSetupType.PROPAGATION_BASED:
        target_device = config.get("target_device", "ANY")
        if target_device != 'NONE':
            hw_config_type = HWConfigType.from_str(HW_CONFIG_TYPE_TARGET_DEVICE_MAP[target_device])

    if isinstance(compression_config_json_section, dict):
        compression_config = NNCFConfig(compression_config_json_section)
        compression_config.register_extra_structs(config.get_all_extra_structs_for_copy())
        compression_config["hw_config_type"] = hw_config_type
        compression_config['quantizer_setup_type'] = quantizer_setup_type
        return [get_compression_algorithm(compression_config)(compression_config, should_init=should_init), ]
    retval = []
    for algo_config in compression_config_json_section:
        algo_config = NNCFConfig(algo_config)
        algo_config.register_extra_structs(config.get_all_extra_structs_for_copy())
        algo_config["hw_config_type"] = hw_config_type
        algo_config['quantizer_setup_type'] = quantizer_setup_type
        retval.append(get_compression_algorithm(algo_config)(algo_config, should_init=should_init))
    return retval


def create_compressed_model(model: Module, config: NNCFConfig,
                            resuming_state_dict: dict = None,
                            dummy_forward_fn: Callable[[Module], Any] = None,
                            wrap_inputs_fn: Callable[[Tuple, Dict], Tuple[Tuple, Dict]] = None,
                            dump_graphs=True,) \
    -> Tuple[CompressionAlgorithmController, NNCFNetwork]:
    """
    The main function used to produce a model ready for compression fine-tuning from an original PyTorch
    model and a configuration object.
    dummy_forward_fn
    :param model: The original model. Should have its parameters already loaded from a checkpoint or another
    source.
    :param config: A configuration object used to determine the exact compression modifications to be applied
    to the model
    :param resuming_state_dict: A PyTorch state dict object to load (strictly) into the compressed model after
    building.
    :param dummy_forward_fn: if supplied, will be used instead of a *forward* function call to build
    the internal graph representation via tracing. Specifying this is useful when the original training pipeline
    has special formats of data loader output or has additional *forward* arguments other than input tensors.
    Otherwise, the *forward* call of the model during graph tracing will be made with mock tensors according
    to the shape specified in the config object.
    :param wrap_inputs_fn: if supplied, will be used on the module's input arguments during a regular, non-dummy
    forward call before passing the inputs to the underlying compressed model. This is required if the model's input
    tensors that are important for compression are not supplied as arguments to the model's forward call directly, but
    instead are located in a container (such as list), and the model receives the container as an argument.
    wrap_inputs_fn should take as input two arguments - the tuple of positional arguments to the underlying
    model's forward call, and a dict of keyword arguments to the same. The function should wrap each tensor among the
    supplied model's args and kwargs that is important for compression (e.g. quantization) with an nncf.nncf_model_input
    function, which is a no-operation function and marks the tensors as inputs to be traced by NNCF in the internal
    graph representation. Output is the tuple of (args, kwargs), where args and kwargs are the same as were supplied in
    input, but each tensor in the original input.
    :param dump_graphs: Whether or not should also dump the internal graph representation of the
    original and compressed models in the .dot format into the log directory.
    :return: A controller for the compression algorithm (or algorithms, in which case the controller
    is an instance of CompositeCompressionController) and the model ready for compression parameter training wrapped
    as an object of NNCFNetwork."""

    if dump_graphs:
        if dummy_forward_fn is None:
            input_info_list = create_input_infos(config)
            graph_builder = GraphBuilder(custom_forward_fn=
                                         create_dummy_forward_fn(input_info_list,
                                                                 with_input_tracing=True))
        else:
            graph_builder = GraphBuilder(custom_forward_fn=dummy_forward_fn)

        if is_main_process():
            graph = graph_builder.build_graph(model)
            graph.visualize_graph(osp.join(config.get("log_dir", "."), "original_graph.dot"))

    if is_debug():
        set_debug_log_dir(config.get("log_dir", "."))

    input_info_list = create_input_infos(config)
    scopes_without_shape_matching = config.get('scopes_without_shape_matching', [])
    ignored_scopes = config.get('ignored_scopes')
    target_scopes = config.get('target_scopes')

    compressed_model = NNCFNetwork(model, input_infos=input_info_list,
                                   dummy_forward_fn=dummy_forward_fn,
                                   wrap_inputs_fn=wrap_inputs_fn,
                                   ignored_scopes=ignored_scopes,
                                   target_scopes=target_scopes,
                                   scopes_without_shape_matching=scopes_without_shape_matching)

    should_init = resuming_state_dict is None
    compression_algo_builder_list = create_compression_algorithm_builders(config, should_init=should_init)

    for builder in compression_algo_builder_list:
        compressed_model = builder.apply_to(compressed_model)
    compression_ctrl = compressed_model.commit_compression_changes()

    if dump_graphs and is_main_process() and compression_algo_builder_list:
        if dummy_forward_fn is None:
            compressed_graph_builder = GraphBuilder(custom_forward_fn=
                                                    create_dummy_forward_fn(input_info_list,
                                                                            with_input_tracing=False))
        else:
            compressed_graph_builder = GraphBuilder(custom_forward_fn=dummy_forward_fn)

        graph = compressed_graph_builder.build_graph(compressed_model, compressed_model.get_tracing_context())
        graph.visualize_graph(osp.join(config.get("log_dir", "."), "compressed_graph.dot"))

    if resuming_state_dict is not None:
        load_state(compressed_model, resuming_state_dict, is_resume=True)

    return compression_ctrl, compressed_model



import warnings

from torch.nn import DataParallel

from nncf.debug import is_debug
from nncf.dynamic_graph.context import get_current_context, OperatorInput
from nncf.dynamic_graph.trace_tensor import flatten_args, trace_tensors
from nncf.layers import ITERATION_MODULES

_IGNORED_SCOPES = []

def _warn_data_parallel():
    if getattr(_warn_data_parallel, 'warned_once', False):
        return
    _warn_data_parallel.warned_once = True
    warnings.warn("You are using DataParallel, which may cause significant performance issues with dynamic graph "
                  "building. Consider using distributed training (DistributedDataParallel) instead")

def ignore_scope(cls):
    if cls not in _IGNORED_SCOPES:
        _IGNORED_SCOPES.append(cls)
    return cls


def wrap_operator(operator, operator_info: 'PatchedOperatorInfo'):
    # do not wrap function twice
    _orig_op = getattr(operator, '_original_op', None)
    if _orig_op is not None:
        raise Exception("Operator: {} is already wrapped".format(_orig_op.__name__))

    def wrapped(*args, **kwargs):
        ctx = get_current_context()
        if not ctx or getattr(ctx, 'in_operator', False) or not ctx.is_tracing:
            op1 = operator(*args, **kwargs)
            return op1

        ctx.in_operator = True

        if operator_info.custom_trace_fn is not None:
            try:
                result = operator_info.custom_trace_fn(operator, *args, **kwargs)
            except:
                # Looks like the __repr__ call made during IDE debug to display tensor contents does not exit properly,
                # but instead throws an exception. This try...except block handles such a situation.
                # Otherwise the context is stuck in the "in_operator == True" state.
                ctx.in_operator = False
                raise
        else:
            ia_op_exec_context = ctx.get_caller_context(operator_info.name)
            ctx.register_operator_call(ia_op_exec_context.operator_name, ia_op_exec_context.scope_in_model)

            op_input = OperatorInput(list(args), kwargs)
            processed_input = ctx.execute_pre_hooks(ia_op_exec_context, op_input)
            args = tuple(processed_input.op_args)
            kwargs = processed_input.op_kwargs
            fargs = flatten_args(args, kwargs)

            node = ctx.find_operator_node(fargs, ia_op_exec_context)
            if is_debug():
                ctx.register_node_call(ctx.graph.get_node_key_by_id(node.node_id))

            result = operator(*args, **kwargs)

            result = trace_tensors(result, node)
            result = ctx.execute_post_hooks(ia_op_exec_context, result)

        ctx.in_operator = False
        return result

    # pylint: disable=protected-access
    wrapped._original_op = operator
    return wrapped


def wrap_module_call(module_call):
    def wrapped(self, *args, **kwargs):
        ctx = get_current_context()
        if not ctx or self.__class__ in _IGNORED_SCOPES:
            if isinstance(self, DataParallel):
                _warn_data_parallel()
            return module_call(self, *args, **kwargs)
        ctx.push_scope(self)
        retval = module_call(self, *args, **kwargs)
        if type(self).__name__ in ITERATION_MODULES.registry_dict.keys():
            ctx.reset_operator_call_count_in_scope(ctx.scope)
        ctx.pop_scope()
        return retval

    return wrapped



"""
 Copyright (c) 2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import networkx as nx
import torch

from nncf.dynamic_graph.graph import NNCFGraph
from nncf.dynamic_graph.operator_metatypes import NoopMetatype, HardTanhMetatype, TanhMetatype, RELUMetatype, \
    PRELUMetatype, ELUMetatype, GELUMetatype, SigmoidMetatype, SoftmaxMetatype, AvgPool2dMetatype, MaxPool2dMetatype, \
    DropoutMetatype, Conv1dMetatype, Conv2dMetatype, Conv3dMetatype, BatchNormMetatype, CatMetatype, AddMetatype, \
    SubMetatype, DivMetatype, MulMetatype, LinearMetatype, MatMulMetatype, MinMetatype, MaxMetatype, MeanMetatype
from nncf.nncf_logger import logger as nncf_logger
from nncf.nncf_network import NNCFNetwork
from nncf.pruning.export_utils import PruningOperationsMetatypeRegistry, identity_mask_propagation, get_input_masks, \
    fill_input_masks
from nncf.pruning.utils import get_sources_of_node

PRUNING_OPERATOR_METATYPES = PruningOperationsMetatypeRegistry("operator_metatypes")


# pylint: disable=protected-access
class DefaultMetaOp:
    subtypes = []
    additional_types = []

    def mask_propagation(self, model: NNCFNetwork, nx_node: dict, graph: NNCFGraph, nx_graph: nx.DiGraph):
        """
        Propagate mask through a node using masks of all inputs and pruning mask of current node (if any).
        Should set the following attributes:
        input_masks - list of masks of input nodes (None if there is no mask in some input)
        output_mask - resulting mask of nx_node operation
        accept_pruned_input - can this operation work with pruned input or not
        :param model: model to prune
        :param nx_node: node from networkx graph to propagate mask through it
        :param graph: graph of model to prune
        :param nx_graph: networkx graph
        """
        raise NotImplementedError

    def input_prune(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        """
        Prune nx_node by input_masks (if masks is not none and operation support it).
        """

    def output_prune(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        """
        Prune nx_node by output_mask (if mask is not none and operation support it).
        """

    @classmethod
    def get_all_op_aliases(cls):
        """
        :return: list of all aliases of types in metatype
        """
        op_types = []
        for subtype in cls.subtypes:
            op_types.extend(subtype.get_all_aliases())
        op_types = list(set(op_types)) + cls.additional_types
        return op_types


@PRUNING_OPERATOR_METATYPES.register('model_input')
class Input(DefaultMetaOp):
    subtypes = [NoopMetatype]

    def mask_propagation(self, model, nx_node, graph, nx_graph):
        nx_node['input_masks'] = []
        nx_node['output_mask'] = None
        nx_node['accept_pruned_input'] = False


@PRUNING_OPERATOR_METATYPES.register('identity_mask_propagation')
class IdentityMaskForwardOps(DefaultMetaOp):
    subtypes = [HardTanhMetatype, TanhMetatype, RELUMetatype, PRELUMetatype, ELUMetatype, GELUMetatype, SigmoidMetatype,
                SoftmaxMetatype, AvgPool2dMetatype, MaxPool2dMetatype, DropoutMetatype]
    additional_types = ["h_sigmoid", "h_swish"]

    def mask_propagation(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        identity_mask_propagation(nx_node, nx_graph)
        nx_node['accept_pruned_input'] = True


@PRUNING_OPERATOR_METATYPES.register('convolution')
class Convolution(DefaultMetaOp):
    subtypes = [Conv1dMetatype, Conv2dMetatype, Conv3dMetatype]

    def mask_propagation(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        output_mask = None
        accept_pruned_input = True
        is_depthwise = False
        input_masks = get_input_masks(nx_node, nx_graph)

        nncf_node = graph._nx_node_to_nncf_node(nx_node)
        node_module = model.get_module_by_scope(nncf_node.op_exec_context.scope_in_model)

        if node_module.pre_ops:
            output_mask = node_module.pre_ops['0'].op.binary_filter_pruning_mask

        # In case of group convs we can't prune by output filters
        if node_module.groups != 1:
            if node_module.weight.size(1) == 1:
                # Depthwise case
                is_depthwise = True
                output_mask = input_masks[0]
            else:
                accept_pruned_input = False
                output_mask = None

        nx_node['input_masks'] = input_masks
        nx_node['output_mask'] = output_mask
        nx_node['accept_pruned_input'] = accept_pruned_input
        nx_node['is_depthwise'] = is_depthwise

    def input_prune(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        input_mask = nx_node['input_masks'][0]
        if input_mask is None:
            return
        bool_mask = torch.tensor(input_mask, dtype=torch.bool)
        new_num_channels = int(torch.sum(input_mask))

        nncf_node = graph._nx_node_to_nncf_node(nx_node)
        node_module = model.get_module_by_scope(nncf_node.op_exec_context.scope_in_model)
        is_depthwise = nx_node['is_depthwise']
        old_num_clannels = int(node_module.weight.size(1))

        if is_depthwise:
            # In depthwise case prune output channels by input mask, here only fix for new number of input channels
            node_module.groups = new_num_channels
            node_module.in_channels = new_num_channels
        else:
            out_channels = node_module.weight.size(0)
            broadcasted_mask = bool_mask.repeat(out_channels).view(out_channels, bool_mask.size(0))
            new_weight_shape = list(node_module.weight.shape)
            new_weight_shape[1] = new_num_channels

            node_module.in_channels = new_num_channels
            node_module.weight = torch.nn.Parameter(node_module.weight[broadcasted_mask].view(new_weight_shape))

        nncf_logger.info('Pruned Convolution {} by input mask. Old input filters number: {}, new filters number:'
                         ' {}.'.format(nx_node['key'], old_num_clannels, new_num_channels))

    def output_prune(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        mask = nx_node['output_mask']
        if mask is None:
            return

        bool_mask = torch.tensor(mask, dtype=torch.bool)

        nncf_node = graph._nx_node_to_nncf_node(nx_node)
        node_module = model.get_module_by_scope(nncf_node.op_exec_context.scope_in_model)
        old_num_clannels = int(node_module.weight.size(0))

        node_module.out_channels = int(torch.sum(mask))
        node_module.weight = torch.nn.Parameter(node_module.weight[bool_mask])

        if node_module.bias is not None and not nx_node['is_depthwise']:
            node_module.bias = torch.nn.Parameter(node_module.bias[bool_mask])

        nncf_logger.info('Pruned Convolution {} by pruning mask. Old output filters number: {}, new filters number:'
                         ' {}.'.format(nx_node['key'], old_num_clannels, node_module.out_channels))


@PRUNING_OPERATOR_METATYPES.register('batch_norm')
class BatchNorm(DefaultMetaOp):
    subtypes = [BatchNormMetatype]

    def mask_propagation(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        identity_mask_propagation(nx_node, nx_graph)
        nx_node['accept_pruned_input'] = True

    def input_prune(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        input_mask = nx_node['input_masks'][0]
        if input_mask is None:
            return

        nncf_node = graph._nx_node_to_nncf_node(nx_node)
        node_module = model.get_module_by_scope(nncf_node.op_exec_context.scope_in_model)

        bool_mask = torch.tensor(input_mask, dtype=torch.bool)
        old_num_clannels = int(node_module.weight.size(0))
        new_num_channels = int(torch.sum(input_mask))

        node_module.num_features = new_num_channels
        node_module.weight = torch.nn.Parameter(node_module.weight[bool_mask])
        node_module.bias = torch.nn.Parameter(node_module.bias[bool_mask])
        node_module.running_mean = torch.nn.Parameter(node_module.running_mean[bool_mask], requires_grad=False)
        node_module.running_var = torch.nn.Parameter(node_module.running_var[bool_mask], requires_grad=False)

        nncf_logger.info('Pruned BatchNorm {} by input mask. Old num features: {}, new num features:'
                         ' {}.'.format(nx_node['key'], old_num_clannels, new_num_channels))


@PRUNING_OPERATOR_METATYPES.register('concat')
class Concat(DefaultMetaOp):
    subtypes = [CatMetatype]

    def all_inputs_from_convs(self, nx_node, nx_graph, graph):
        """
        Return whether all input sources of nx_node is convolutions or not
        :param nx_node: node to determine it's sources
        :param nx_graph:  networkx graph to work with
        :param graph:  NNCF graph to work with
        """
        inputs = [u for u, _ in nx_graph.in_edges(nx_node['key'])]
        input_masks = get_input_masks(nx_node, nx_graph)

        for i, inp in enumerate(inputs):
            # If input has mask ->  it went from convolution (source of this node is a convolution)
            if input_masks[i] is not None:
                continue
            nncf_input_node = graph._nx_node_to_nncf_node(nx_graph.nodes[inp])
            source_nodes = get_sources_of_node(nncf_input_node, graph, Convolution.get_all_op_aliases() +
                                               StopMaskForwardOps.get_all_op_aliases() +
                                               Input.get_all_op_aliases())
            sources_types = [node.op_exec_context.operator_name for node in source_nodes]
            if any([t in sources_types for t in StopMaskForwardOps.get_all_op_aliases()]):
                return False
        return True

    def check_concat(self, nx_node, nx_graph, graph):
        if self.all_inputs_from_convs(nx_node, nx_graph, graph):
            return True
        return False

    def mask_propagation(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        result_mask = None
        accept_pruned_input = False

        if self.check_concat(nx_node, nx_graph, graph):
            input_masks, filled_input_masks = fill_input_masks(nx_node, nx_graph)

            if all([mask is None for mask in input_masks]):
                result_mask = None
            else:
                result_mask = torch.cat(filled_input_masks)
                accept_pruned_input = True

        nx_node['input_masks'] = input_masks
        nx_node['output_mask'] = result_mask
        nx_node['accept_pruned_input'] = accept_pruned_input


@PRUNING_OPERATOR_METATYPES.register('elementwise')
class Elementwise(DefaultMetaOp):
    subtypes = [AddMetatype, SubMetatype, DivMetatype, MulMetatype]

    def mask_propagation(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        input_masks = get_input_masks(nx_node, nx_graph)

        nx_node['input_masks'] = input_masks
        nx_node['output_mask'] = None
        nx_node['accept_pruned_input'] = False


@PRUNING_OPERATOR_METATYPES.register('stop_propagation_ops')
class StopMaskForwardOps(DefaultMetaOp):
    subtypes = [MeanMetatype, MaxMetatype, MinMetatype, LinearMetatype, MatMulMetatype]

    def mask_propagation(self, model: NNCFNetwork, nx_node, graph: NNCFGraph, nx_graph: nx.DiGraph):
        input_masks = get_input_masks(nx_node, nx_graph)

        nx_node['input_masks'] = input_masks
        nx_node['output_mask'] = None
        nx_node['accept_pruned_input'] = False


class ModelPruner:
    CAN_PRUNE_ATTR = 'can_prune'

    def __init__(self, model: NNCFNetwork, graph: NNCFGraph, nx_graph: nx.DiGraph):
        self.model = model
        self.graph = graph
        self.nx_graph = nx_graph

    def node_propagate_can_prune_attr(self, nx_node_key):
        nx_node = self.nx_graph.nodes[nx_node_key]
        is_conv = self.graph.node_type_fn(nx_node) in Convolution().get_all_op_aliases()
        return not is_conv or (is_conv and nx_node['is_depthwise'])

    def node_accept_different_inputs(self, nx_node):
        """
        Return whether nx_node accept pruned and not pruned inputs as inputs at the same time.
        """
        node_type = self.graph.node_type_fn(nx_node)
        return node_type in Concat.get_all_op_aliases()

    @staticmethod
    def get_class_by_type_name(type_name):
        """
        Return class of metaop that corresponds to type_name type.
        """
        cls = PRUNING_OPERATOR_METATYPES.get_operator_metatype_by_op_name(type_name)
        if cls is None:
            nncf_logger.warning(
                "Layer {} is not pruneable - will not propagate pruned filters through it".format(type_name))
            cls = StopMaskForwardOps
        return cls

    def mask_propagation(self):
        """
        Mask propagation in graph:
        to propagate masks run method mask_propagation (of metaop of current node) on all nodes in topological order.
        """
        sorted_nodes = [self.nx_graph.nodes[node_name] for node_name in nx.topological_sort(self.nx_graph)]
        for node in sorted_nodes:
            node_type = self.graph.node_type_fn(node)
            cls = self.get_class_by_type_name(node_type)()
            cls.mask_propagation(self.model, node, self.graph, self.nx_graph)
        nncf_logger.info('Finished mask propagation in graph')

    def propagate_can_prune_attr_up(self):
        """
        Propagating can_prune attribute in reversed topological order.
        This attribute depends on accept_pruned_input and can_prune attributes of output nodes.
        Node can_prune is True if all outputs accept_pruned_input is True and all outputs
        (except convs because conv can be pruned by input and output independently).
        """
        for node_name in self.nx_graph.nodes:
            self.nx_graph.nodes[node_name][ModelPruner.CAN_PRUNE_ATTR] = True

        reversed_sorted_nodes = reversed([self.nx_graph.nodes[name] for name in nx.topological_sort(self.nx_graph)])
        for node in reversed_sorted_nodes:
            # Check all output nodes accept_pruned_input attribute
            out_edges = self.nx_graph.out_edges(node['key'])
            outputs_accept_pruned_input = all(self.nx_graph.nodes[key]['accept_pruned_input'] for _, key in out_edges)

            # Check all output nodes can_prune attribute
            outputs_will_be_pruned = all([self.nx_graph.nodes[key][ModelPruner.CAN_PRUNE_ATTR] for _, key in out_edges
                                          if self.node_propagate_can_prune_attr(key)])
            node[ModelPruner.CAN_PRUNE_ATTR] = outputs_accept_pruned_input and outputs_will_be_pruned

        nncf_logger.info('Propagated can_prune attribute up')

    def propagate_can_prune_attr_down(self):
        """
        Propagating can_prune attribute down to fix all branching cases with one pruned and one not pruned
        branches.
        """
        sorted_nodes = [self.nx_graph.nodes[name] for name in nx.topological_sort(self.nx_graph)]
        for node in sorted_nodes:
            # Propagate attribute only in not conv case
            if self.node_propagate_can_prune_attr(node['key']):
                in_edges = self.nx_graph.in_edges(node['key'])
                can_prune = all(self.nx_graph.nodes[key][ModelPruner.CAN_PRUNE_ATTR] for key, _ in in_edges)
                can_prune_any = any(self.nx_graph.nodes[key][ModelPruner.CAN_PRUNE_ATTR] for key, _ in in_edges)

                if (not self.node_accept_different_inputs(node) and not can_prune) or \
                        (self.node_accept_different_inputs(node) and not can_prune_any):
                    node[ModelPruner.CAN_PRUNE_ATTR] = can_prune

        nncf_logger.info('Propagated can_prune attribute down')

    def apply_mask(self):
        """
        Applying propagated masks for all nodes in topological order:
        if all inputs of node can_prune -> running input_prune method for this node
        if node[ModelPruner.CAN_PRUNE_ATTR] -> running output_prune method for this node
        """
        sorted_nodes = [self.nx_graph.nodes[name] for name in nx.topological_sort(self.nx_graph)]
        with torch.no_grad():
            for node in sorted_nodes:
                node_type = self.graph.node_type_fn(node)
                node_cls = self.get_class_by_type_name(node_type)()

                in_edges = self.nx_graph.in_edges(node['key'])
                can_prune_input = all(self.nx_graph.nodes[key][ModelPruner.CAN_PRUNE_ATTR] for key, _ in in_edges)
                if can_prune_input:
                    node_cls.input_prune(self.model, node, self.graph, self.nx_graph)

                if node[ModelPruner.CAN_PRUNE_ATTR]:
                    node_cls.output_prune(self.model, node, self.graph, self.nx_graph)
        nncf_logger.info('Finished mask applying step')

    def prune_model(self):
        """
        Model pruner work in three stages:
        1. Mask propagation: propagate pruning masks through the graph.
        2. Propagate can_prune attribute (up and then down) through the graph.This attribute shows can we really
        prune some node or not.
        3. Applying masks accordingly with can_prune attribute (only when can prune).
        :return:
        """
        nncf_logger.info('Start pruning model')
        self.mask_propagation()
        self.propagate_can_prune_attr_up()
        self.propagate_can_prune_attr_down()
        self.apply_mask()
        nncf_logger.info('Finished pruning model')



"""
 Copyright (c) 2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import itertools
from collections import OrderedDict
from enum import Enum
from typing import List, Dict, Union, Tuple, NamedTuple, Callable, Any

import torch
import warnings
from bisect import bisect_left
from operator import itemgetter
from torch import Tensor, nn
from torch.nn.modules.loss import _Loss

from nncf.debug import is_debug
from nncf.dynamic_graph.context import Scope
from nncf.nncf_logger import logger as nncf_logger
from nncf.quantization.layers import QuantizersSwitcher
from .adjacent_quantizers import GroupsOfAdjacentQuantizers, AdjacentQuantizers
from .compression_ratio import CompressionRatioCalculator
from .hawq_debug import HAWQDebugger
from .manual_init import ManualPrecisionInitializer
from .perturbations import Perturbations, PerturbationObserver
from .traces_order import TracesPerLayer, TracesOrder
from ..hessian_trace import HessianTraceEstimator
from ..hw_precision_constraints import HWPrecisionConstraints
from ..quantizer_id import QuantizerId
from ...layer_utils import ProxyModule
from ...module_operations import UpdateParameter
from ...structures import QuantizationPrecisionInitArgs


class BitwidthAssignmentMode(Enum):
    STRICT = 'strict'
    LIBERAL = 'liberal'

    @staticmethod
    def from_str(config_value: str) -> 'BitwidthAssignmentMode':
        if config_value == BitwidthAssignmentMode.STRICT.value:
            return BitwidthAssignmentMode.STRICT
        if config_value == BitwidthAssignmentMode.LIBERAL.value:
            return BitwidthAssignmentMode.LIBERAL
        raise RuntimeError("Unknown bitwidth assignment mode")


class HAWQPrecisionInitializer(ManualPrecisionInitializer):
    def __init__(self, algo: 'QuantizationController', config: 'NNCFConfig',
                 init_args: QuantizationPrecisionInitArgs):
        super().__init__(algo, config, init_args)
        self._criterion_fn = init_args.criterion_fn
        self._criterion = init_args.criterion
        self._data_loader = init_args.data_loader
        self._traces_per_layer_path = config.get('traces_per_layer_path', None)
        self._num_data_points = config.get('num_data_points', 1000)
        self._iter_number = config.get('iter_number', 500)
        self._tolerance = config.get('tolerance', 1e-5)
        self._compression_ratio = config.get('compression_ratio', 1.5)
        self._bits = self._hw_precision_constraints.get_all_unique_bits() \
            if self._hw_precision_constraints else config.get('bits', [4, 8])
        self._init_device = init_args.device
        self.flops_counter = CompressionRatioCalculator(self._model, self._quantizers_handler)
        self._groups_of_adjacent_quantizers = GroupsOfAdjacentQuantizers(algo)
        self._dump_hawq_data = config.get('dump_hawq_data', False)
        bitwidth_assignment_mode_str = config.get('bitwidth_assignment_mode', BitwidthAssignmentMode.LIBERAL.value)
        self._bitwidth_assignment_mode = BitwidthAssignmentMode.from_str(bitwidth_assignment_mode_str)

    def apply_init(self):
        if not self._quantizers_handler.get_weight_quantizers_in_execution_order_per_id():
            return None
        original_device = next(self._model.parameters()).device
        self._model.to(self._init_device)

        traces_per_layer = self._calc_traces(self._criterion_fn, self._criterion, self._iter_number, self._tolerance)
        if not traces_per_layer:
            raise RuntimeError('Failed to calculate hessian traces!')

        traces_order = traces_per_layer.traces_order
        num_weights = len(self._weight_quantizations_by_execution_order)
        bits_configurations = self.get_configs_constrained_by_traces_order(self._bits, num_weights)

        weight_quantizer_ids_in_execution_order = list(self._weight_quantizations_by_execution_order.keys())

        if self._bitwidth_assignment_mode == BitwidthAssignmentMode.STRICT:
            self._merge_constraints_for_adjacent_quantizers(self._groups_of_adjacent_quantizers,
                                                            self._hw_precision_constraints)

        bits_configurations = self._filter_configs_by_precision_constraints(bits_configurations,
                                                                            self._hw_precision_constraints,
                                                                            weight_quantizer_ids_in_execution_order,
                                                                            traces_order)
        if not bits_configurations:
            warnings.warn('All bits configurations are incompatible with HW Config!', RuntimeWarning)
            return None

        if self._bitwidth_assignment_mode == BitwidthAssignmentMode.STRICT:
            bits_configurations = \
                self._filter_configs_by_grouped_weight_quantizers(bits_configurations,
                                                                  weight_quantizer_ids_in_execution_order,
                                                                  self._groups_of_adjacent_quantizers,
                                                                  traces_order)
        if not bits_configurations:
            warnings.warn('No bits configurations are left after removing inconsistent groups of weight quantizers'
                          ' with adjacent activation quantizers!', RuntimeWarning)
            return None

        flops_bits_per_config = self.get_flops_bits_per_config(bits_configurations, traces_order)
        min_ratio = min(flops_bits_per_config)
        max_ratio = max(flops_bits_per_config)
        if not min_ratio <= self._compression_ratio <= max_ratio:
            raise AttributeError('Invalid compression ratio={}. Should be within range [{:.3f}, {:.3f}]'.format(
                self._compression_ratio, min_ratio, max_ratio))

        perturbations, weight_observers = self.calc_quantization_noise()

        configuration_metric = self.calc_hawq_metric_per_configuration(bits_configurations, perturbations,
                                                                       traces_per_layer, self._init_device)

        config_index = self.choose_configuration(configuration_metric, flops_bits_per_config)
        chosen_config_in_traces_order = bits_configurations[config_index]
        chosen_config_in_execution_order = traces_order.get_execution_order_config(chosen_config_in_traces_order)
        nncf_logger.info('Chosen HAWQ configuration with ratio={:.2f}, bitwidth per weightable layer={}'.format(
            flops_bits_per_config[config_index], chosen_config_in_execution_order))
        nncf_logger.debug('Order of the weightable layers in the HAWQ configuration (in descending order of average '
                          'Hessian traces) ={}'.format(traces_order))

        self.set_chosen_config(chosen_config_in_execution_order)
        self._model.rebuild_graph()
        if is_debug() or self._dump_hawq_data:
            hawq_debugger = HAWQDebugger(bits_configurations, perturbations,
                                         weight_observers, traces_per_layer, self._bits)
            hawq_debugger.dump_metric_MB(configuration_metric)
            hawq_debugger.dump_metric_flops(configuration_metric, flops_bits_per_config, config_index)
            hawq_debugger.dump_avg_traces()
            hawq_debugger.dump_density_of_quantization_noise()
            hawq_debugger.dump_perturbations_ratio()
            hawq_debugger.dump_bitwidth_graph(self._algo, self._model, self._groups_of_adjacent_quantizers)
        str_bw = [str(element) for element in self.get_bitwidth_per_scope()]
        nncf_logger.info('\n'.join(['\n\"bitwidth_per_scope\": [', ',\n'.join(str_bw), ']']))

        self._model.to(original_device)

        ordered_metric_per_layer = self.get_metric_per_layer(chosen_config_in_execution_order, perturbations,
                                                             traces_per_layer)
        return ordered_metric_per_layer

    @staticmethod
    def _merge_constraints_for_adjacent_quantizers(groups_of_adjacent_quantizers, hw_precision_constraints):
        if not hw_precision_constraints:
            return
        for group in groups_of_adjacent_quantizers:
            all_bits_sets = []
            quantizer_ids = []
            all_quantizers = group.weight_quantizers + group.activation_quantizers
            for quantizer_id, _ in all_quantizers:
                all_bits_sets.append(hw_precision_constraints.get(quantizer_id))
                quantizer_ids.append(quantizer_id)
            minimal_set_bits = set.intersection(*all_bits_sets)
            for quantizer_id in quantizer_ids:
                if not minimal_set_bits:
                    raise RuntimeError(
                        'No bits configurations are left after removing inconsistent groups of weight quantizers'
                        ' with adjacent activation quantizers!')
                hw_precision_constraints.replace(quantizer_id, minimal_set_bits)

    def get_flops_bits_per_config(self, bits_configurations: List[List[int]], traces_order: TracesOrder) -> List[float]:
        skipped = self._quantizers_handler.get_skipped_weight_quantizers_per_id()
        flops_bits_per_config = []
        for bits_config in bits_configurations:
            execution_order_config = traces_order.get_execution_order_config(bits_config)
            flops_bits_per_config.append(
                self.flops_counter.ratio_for_bits_configuration(execution_order_config, skipped))
        return flops_bits_per_config

    def get_bitwidth_per_scope(self) -> List[List[Union[int, str]]]:
        sorted_quantizers = OrderedDict(sorted(self._all_quantizers_per_scope.items(), key=lambda x: str(x[0])))
        full_bitwidth_per_scope = []
        for scope, quantizer in sorted_quantizers.items():
            full_bitwidth_per_scope.append([quantizer.num_bits, str(scope)])
        return full_bitwidth_per_scope

    class ParamsToRestore(NamedTuple):
        originally_disabled_gradients: List[str]
        skipped_gradients_to_enable: List[Tuple[nn.Module, str]]

    @staticmethod
    def disable_all_gradients_except_weights_of_quantized_modules(
            quantizers_switcher: QuantizersSwitcher,
            quantized_weight_modules_registry: Dict[str, torch.nn.Module],
            model: nn.Module,
            scopes_of_skipped_weight_quantizers: List[
                'Scope'] = None) -> ParamsToRestore:  # pylint: disable=undefined-variable
        """
        Disables gradients of all parameters, except for layers that have quantizers for weights, which wasn't skipped
        because of single precision constraints.
        :param quantizers_switcher: object that is responsible for enabling and disabling quantizers
        :param quantized_weight_modules_registry: modules with quantized weights per scope
        :param model: model to access all parameters
        :param scopes_of_skipped_weight_quantizers: list of string scopes of layers that have a single precision
        constraint and which weights should be skipped from bitwidth initialization
        :return: list of names of the parameters that were originally disabled
        """
        originally_disabled_gradients = []
        skipped_gradients_to_enable = []

        # Some quantizers can be disabled in a staged scenario on creation of staged scheduler
        # Need to save originally disabled quantizers for restoring their state after initialization
        quantizers_switcher.disable_quantizers()

        # remember gradients of quantized modules that were enabled
        gradients_to_enable = []
        for scope, quantized_module in quantized_weight_modules_registry.items():
            is_skipped = False
            for skipped_weight_quantizer_scope in scopes_of_skipped_weight_quantizers:
                if skipped_weight_quantizer_scope in Scope.from_str(scope):
                    is_skipped = True
                    break
            for param_name, param in quantized_module.named_parameters():
                if param.requires_grad:
                    # disable gradients for skipped module for optimization of Hessian Trace search
                    if is_skipped:
                        skipped_gradients_to_enable.append((quantized_module, param_name))
                        param.requires_grad = False
                    else:
                        gradients_to_enable.append((quantized_module, param_name))

        # disable all gradients, except already disabled
        for param_name, param in model.named_parameters():
            if not param.requires_grad:
                originally_disabled_gradients.append(param_name)
            else:
                param.requires_grad = False

        # enable gradients of quantized modules that were disabled
        for quantized_module in quantized_weight_modules_registry.values():
            for param_name, param in quantized_module.named_parameters():
                if (quantized_module, param_name) in gradients_to_enable and not 'bias' in param_name:
                    param.requires_grad = True
        return HAWQPrecisionInitializer.ParamsToRestore(originally_disabled_gradients, skipped_gradients_to_enable)

    def _calc_traces(self, criterion_fn: Callable[[Any, Any, _Loss], torch.Tensor], criterion: _Loss,
                     iter_number: int, tolerance: float) -> TracesPerLayer:
        if self._traces_per_layer_path:
            return TracesPerLayer(torch.load(self._traces_per_layer_path).to(self._init_device))

        quantizers_switcher = QuantizersSwitcher(list(self._all_quantizers_per_scope.values()))
        params_to_restore = self.disable_all_gradients_except_weights_of_quantized_modules(
            quantizers_switcher,
            self._algo.quantized_weight_modules_registry,
            self._model,
            self._quantizers_handler.get_scope_of_skipped_weight_quantizers())

        trace_estimator = HessianTraceEstimator(self._model, criterion_fn, criterion, self._init_device,
                                                self._data_loader, self._num_data_points)
        avg_traces = trace_estimator.get_average_traces(max_iter=iter_number, tolerance=tolerance)

        self.restore_disabled_gradients(quantizers_switcher, self._model, self._algo.quantized_weight_modules_registry,
                                        params_to_restore)

        return TracesPerLayer(avg_traces)

    @staticmethod
    def restore_disabled_gradients(quantizers_switcher: QuantizersSwitcher,
                                   model: nn.Module,
                                   quantized_weight_modules_registry: Dict[str, torch.nn.Module],
                                   params_to_restore: ParamsToRestore):
        """
        Restore requires_grad property of all parameters back, except for ones that were originally disabled
        :param quantizers_switcher: object that is responsible for enabling and disabling quantizers
        :param model: model to access all parameters
        :param quantized_weight_modules_registry: modules with quantized weights per scope
        :param params_to_restore: storage names of the parameters that should restore reguires_grad property
        """
        for quantized_module in quantized_weight_modules_registry.values():
            for param_name, param in quantized_module.named_parameters():
                if (quantized_module, param_name) in params_to_restore.skipped_gradients_to_enable:
                    param.requires_grad = True

        for param_name, param in model.named_parameters():
            if param_name not in params_to_restore.originally_disabled_gradients:
                param.requires_grad = True
        quantizers_switcher.enable_quantizers()

    @staticmethod
    def get_configs_constrained_by_traces_order(bits_: List[int], num_layers: int) -> List[List[int]]:
        bit_configs = []
        if num_layers == 0:
            return bit_configs
        bits = sorted(bits_)
        m = len(bits)
        L = num_layers
        for j in range(1, m + 1):
            for combo_bits in itertools.combinations(bits, j):
                for combo_partitions in itertools.combinations(list(range(1, L)), j - 1):
                    bit_config = []
                    prev_p = 0
                    for (p, b) in zip(combo_partitions + (L,), combo_bits):
                        bit_config += [b] * (p - prev_p)
                        prev_p = p
                    bit_configs.append(bit_config)
        return bit_configs

    @staticmethod
    def _filter_configs_by_precision_constraints(bits_configurations: List[List[int]],
                                                 hw_precision_constraints: HWPrecisionConstraints,
                                                 ordered_weight_ids: List[QuantizerId],
                                                 traces_order: TracesOrder) -> List[List[int]]:
        if not hw_precision_constraints:
            return bits_configurations

        filtered_bits_configurations = []
        for bits_configuration in bits_configurations:
            is_all_bitwidth_compatible = True
            ordered_config = traces_order.get_execution_order_config(bits_configuration)
            for i, bitwidth in enumerate(ordered_config):
                weight_id = ordered_weight_ids[i]
                bits_constraints = hw_precision_constraints.get(weight_id)
                if bitwidth not in bits_constraints:
                    is_all_bitwidth_compatible = False
                    break
            if is_all_bitwidth_compatible:
                filtered_bits_configurations.append(bits_configuration)
        return filtered_bits_configurations

    def calc_quantization_noise(self) -> [Perturbations, List[PerturbationObserver]]:
        hook_handles = []
        observers = []
        for module in self._weight_quantizations_by_execution_order.values():
            observer = PerturbationObserver(self._init_device)
            hook_handles.append(module.register_forward_hook(observer.calc_perturbation))
            observers.append(observer)

        perturbations = Perturbations()
        for b in self._bits:
            for wi in self._weight_quantizations_by_execution_order.values():
                wi.num_bits = b

            # TODO: replace with do_dummy_forward call on compressing in eval mode only
            # Call each UpdateWeight op, instead of calling dummy_forward. It's needed because dummy_forward must be
            # run with force_eval=False, which overrides BatchNorm statistics. This requirement comes from the models
            # with quantizers on the branches, which are enabled in train mode (AuxLogits for Inception3)
            for quantized_module in self._algo.quantized_weight_modules_registry.values():
                ops = [op for op in quantized_module.pre_ops.values() if isinstance(op, UpdateParameter)]
                ops += [op for op in quantized_module.post_ops.values() if isinstance(op, UpdateParameter)]
                for op in ops:
                    op(ProxyModule(quantized_module), None)

            for i, observer in enumerate(observers):
                perturbations.add(layer_id=i, bitwidth=b, perturbation=observer.get_observation().to(self._init_device))

        for handle in hook_handles:
            handle.remove()
        return perturbations, observers

    @staticmethod
    def calc_hawq_metric_per_configuration(bits_configurations: List[List[int]], perturbations: Perturbations,
                                           traces_per_layer: TracesPerLayer, device) -> List[Tensor]:
        configuration_metric = []
        for bits_config in bits_configurations:
            hawq_metric = torch.Tensor([0]).to(device)
            for trace_index, layer_bits in enumerate(bits_config):
                execution_index = traces_per_layer.traces_order.get_execution_index_by_traces_index(trace_index)
                hawq_metric += traces_per_layer.get_by_trace_index(trace_index) * perturbations.get(
                    layer_id=execution_index, bitwidth=layer_bits)
            configuration_metric.append(hawq_metric)
        return configuration_metric

    def choose_configuration(self, configuration_metric: List[Tensor], flops_bits_per_config: List[float]) -> int:
        num_configs = len(configuration_metric)

        sorted_flops_order = [x[0] for x in sorted(enumerate(flops_bits_per_config), reverse=False, key=lambda x: x[1])]
        sorted_flops_bits_per_config = sorted(flops_bits_per_config)

        boundary_index = bisect_left(sorted_flops_bits_per_config, self._compression_ratio)
        indexes_to_check = [sorted_flops_order[i] for i in range(boundary_index, num_configs)]
        best_metric = min(list(itemgetter(*indexes_to_check)(configuration_metric)))
        best_config_index = configuration_metric.index(best_metric)
        return best_config_index

    def set_chosen_config(self, weight_bitwidth_in_execution_order: List[int]):
        for wq, bits in zip(self._weight_quantizations_by_execution_order.values(), weight_bitwidth_in_execution_order):
            wq.num_bits = bits
        if self._groups_of_adjacent_quantizers:
            for group in self._groups_of_adjacent_quantizers:
                weight_bitwidth_set = {wq.num_bits for _, wq in group.weight_quantizers}
                if self._bitwidth_assignment_mode == BitwidthAssignmentMode.STRICT:
                    self._set_activations_bitwidth_strictly(group, weight_bitwidth_set)
                else:
                    self._set_activation_bitwidth_liberally(group, weight_bitwidth_set)
        else:
            # TODO: delete not-consistent pairs of activation and weights for pattern-based approach
            pairs = self._algo.get_weights_activation_quantizers_pairs()
            for pair in pairs:
                wqs, aq = pair
                aq.num_bits = max([wq.num_bits for wq in wqs])

    def _set_activation_bitwidth_liberally(self, group: AdjacentQuantizers, weight_bitwidth_set):
        for quantizer_id, aq in group.activation_quantizers:
            activation_bitwidth_set = self._hw_precision_constraints.get(quantizer_id)
            intersection = activation_bitwidth_set.intersection(weight_bitwidth_set)
            if activation_bitwidth_set.__len__() == 1:
                aq.num_bits = activation_bitwidth_set.pop()
            elif intersection:
                aq.num_bits = min(intersection)
            elif activation_bitwidth_set:
                aq.num_bits = min(activation_bitwidth_set)
            elif weight_bitwidth_set:
                aq.num_bits = min(weight_bitwidth_set)

    def _set_activations_bitwidth_strictly(self, group: AdjacentQuantizers, weight_bitwidth_set):
        if len(weight_bitwidth_set) > 1:
            raise RuntimeError('Invalid grouping of weight quantizers')
        all_constraints = set()
        for quantizer_id, aq in group.activation_quantizers:
            all_constraints.update(self._hw_precision_constraints.get(quantizer_id))
        common_constraints = set(all_constraints)
        for quantizer_id, aq in group.activation_quantizers:
            constraint = self._hw_precision_constraints.get(quantizer_id)
            common_constraints = common_constraints.intersection(constraint)
        if weight_bitwidth_set:
            common_constraints = common_constraints.intersection(weight_bitwidth_set)
        if not common_constraints:
            raise RuntimeError('No hardware compatible bitwidth for activation quantizers')
        for quantizer_id, aq in group.activation_quantizers:
            aq.num_bits = sorted(list(common_constraints))[0]

    def get_metric_per_layer(self, chosen_config_in_execution_order: List[int], perturbations: Perturbations,
                             traces_per_layer: TracesPerLayer):
        metric_per_layer = []
        for i, layer_bits in enumerate(chosen_config_in_execution_order):
            metric_per_layer.append(traces_per_layer.get_by_execution_index(i) * perturbations.get(i, layer_bits))
        ordered_metric_per_layer = [i[0] for i in
                                    sorted(enumerate(metric_per_layer), reverse=True, key=lambda x: x[1])]
        return ordered_metric_per_layer

    @staticmethod
    def _filter_configs_by_grouped_weight_quantizers(bits_configurations: List[List[int]],
                                                     weight_quantization_ids_by_execution_order: List[QuantizerId],
                                                     groups_of_adjacent_quantizers: GroupsOfAdjacentQuantizers,
                                                     traces_order: TracesOrder) -> List[List[int]]:
        """ removes configs where adjacent weight quantizers have different bitwidth. Adjacency is defined by common
        activation quantizers"""
        filtered_bits_configurations = []
        all_grouped_indexes = []
        for group_of_adjacent_quantizers in groups_of_adjacent_quantizers:
            wqs = group_of_adjacent_quantizers.weight_quantizers
            if len(wqs) > 1:
                indexes_of_grouped_wq = []
                for quantizer_id, _ in wqs:
                    index_by_execution_order = weight_quantization_ids_by_execution_order.index(quantizer_id)
                    indexes_of_grouped_wq.append(index_by_execution_order)
                all_grouped_indexes.append(indexes_of_grouped_wq)

        if not all_grouped_indexes:
            return bits_configurations

        for bits_configuration in bits_configurations:
            bitwidth_by_execution_order = traces_order.get_execution_order_config(bits_configuration)
            keep_config = True
            for indexes_of_grouped_wq in all_grouped_indexes:
                grouped_bits = [bitwidth_by_execution_order[index] for index in indexes_of_grouped_wq]
                if grouped_bits[1:] != grouped_bits[:-1]:
                    keep_config = False
                    break
            if keep_config:
                filtered_bits_configurations.append(bits_configuration)

        return filtered_bits_configurations



"""
 Copyright (c) 2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
from collections import OrderedDict
from typing import List, Tuple, NamedTuple

from nncf.quantization.layers import BaseQuantizer
from ..quantizer_id import QuantizerId
from ..quantizer_propagation import QuantizersBetweenQuantizableLayers


class AdjacentQuantizers(NamedTuple):
    activation_quantizers: List[Tuple[QuantizerId, BaseQuantizer]]
    weight_quantizers: List[Tuple[QuantizerId, BaseQuantizer]]


class GroupsOfAdjacentQuantizers:
    def __init__(self, quantization_ctrl: 'QuantizationController'):
        repeated_groups = []
        non_weight_quantizers = quantization_ctrl.non_weight_quantizers
        sorted_quantizers = OrderedDict(sorted(non_weight_quantizers.items(), key=lambda x: str(x[0])))
        for quantizer_id, quantizer_info in sorted_quantizers.items():
            group = quantizer_info.quantizers_between_quantizable_layers  # type: QuantizersBetweenQuantizableLayers
            if group:
                repeated_groups.append(group)

        self._quantizer_per_group_id = {}
        self._groups_of_adjacent_quantizers: List[AdjacentQuantizers] = []

        unique_groups = list(dict.fromkeys(repeated_groups))

        for i, group in enumerate(unique_groups):
            quantized_module_scopes = group.quantized_module_scopes
            paired_wq = []
            for scope in quantized_module_scopes:
                for quantizer_id, quantizer in quantization_ctrl.weight_quantizers.items():
                    if scope == quantizer_id.get_scope():
                        paired_wq.append((quantizer_id, quantizer))
                        self._quantizer_per_group_id[id(quantizer)] = i
                        break
            paired_aq = []
            for ia_op_ctx in group.activation_quantizer_ctxs:
                for quantizer_id, quantizer_info in quantization_ctrl.non_weight_quantizers.items():
                    if ia_op_ctx == quantizer_id.ia_op_exec_context:
                        quantizer = quantizer_info.quantizer_module_ref
                        paired_aq.append((quantizer_id, quantizer))
                        self._quantizer_per_group_id[id(quantizer)] = i
                        break

            self._groups_of_adjacent_quantizers.append(AdjacentQuantizers(paired_aq, paired_wq))

    def get_group_id_for_quantizer(self, quantizer: BaseQuantizer):
        qid = id(quantizer)
        return self._quantizer_per_group_id.get(qid, None)

    def __iter__(self):
        return iter(self._groups_of_adjacent_quantizers)

    def __bool__(self):
        return bool(self._groups_of_adjacent_quantizers) and bool(self._quantizer_per_group_id)



"""
 Copyright (c) 2019 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from itertools import chain, combinations
from typing import Callable, List

import numpy as np
import networkx as nx


def powerset(iterable, min_r=1, max_r=None):
    if not isinstance(iterable, list):
        s = list(iterable)
    else:
        s = iterable
    if max_r is None:
        max_r = len(s)
    return chain.from_iterable(combinations(s, r) for r in range(min_r, max_r + 1))


class Expression:
    def _match(self, nodes, graph):
        return NotImplementedError

    def __add__(self, other):
        return ConcatExpression([self, other])

    def __or__(self, other):
        return AlternatingExpression([self, other])

    def __and__(self, other):
        return BranchingExpression([self, other])

    def _iterate_alternatives(self, nodes):
        return powerset(nodes, min_r=1)

    def match(self, nodes, graph):
        all_matches = []
        for n in self._iterate_alternatives(nodes):
            result = self._match(n, graph)
            if not result:
                continue

            n, following = result
            following = list(following)
            if not isinstance(n, list):
                n = [n]

            all_matches.append((n, following))
        if not all_matches:
            return None, None
        return max(all_matches, key=lambda x: len(x[0]))


class ConcatExpression(Expression):
    def __init__(self, expressions):
        self.expressions = expressions

    def _match(self, nodes, graph):
        assert len(self.expressions) > 1
        full_match = []
        for ex in self.expressions:
            matches, following = ex.match(nodes, graph)

            if not matches:
                return None

            full_match += matches

            nodes = following
        return full_match, following

    def __add__(self, other):
        return ConcatExpression(self.expressions + [other])


class AlternatingExpression(Expression):
    def __init__(self, expressions, greedy_match=False, greedy_consume=True):
        self.greedy_match = greedy_match
        self.greedy_consume = greedy_consume
        self.expressions = expressions

    def _match(self, nodes, graph):
        assert len(self.expressions) > 1
        all_matches = []
        for ex in self.expressions:
            matched, following = ex.match(nodes, graph)
            if not matched:
                continue

            if self.greedy_match:
                return matched, following

            all_matches.append((matched, following))

        if self.greedy_consume:
            if not all_matches:
                return None
            return max(all_matches, key=lambda x: len(x[0]))
        return None

    def __or__(self, other):
        return AlternatingExpression(self.expressions + [other])


class BranchingExpression(Expression):
    def __init__(self, expressions):
        self.expressions = expressions

    def _iterate_alternatives(self, nodes):
        return powerset(nodes, len(self.expressions), len(self.expressions))

    def _match(self, nodes, graph):
        assert len(self.expressions) > 1
        if len(nodes) != len(self.expressions):
            # need to try all possible assignments
            return None

        matches = [[] for _ in range(len(self.expressions))]
        for i, ex in enumerate(self.expressions):
            any_matched = False
            for node_name in nodes:
                matched, following = ex.match([node_name], graph)
                matches[i].append((matched, following))

                if matched:
                    any_matched = True

            if not any_matched:
                return None

        return self._assign_matches(matches)

    def _assign_matches(self, matches):
        """Assign every expression to some match"""
        assignments = np.full(len(matches[0]), -1)
        used = np.full(len(matches), False)

        def _find_assignment(i):
            if used[i]:
                return False
            used[i] = True
            for j in range(len(matches[0])):
                if not matches[i][j][0]:
                    continue
                if assignments[j] == -1 or _find_assignment(assignments[j]):
                    assignments[j] = i
                    return True
            return False

        for i in range(len(self.expressions)):
            used[...] = False
            _find_assignment(i)

        all_matches = set()
        all_followings = set()
        for i in range(len(matches[0])):
            if assignments[i] != -1:
                match, follow = matches[assignments[i]][i]
                all_matches.update(match)
                all_followings.update(follow)

        # assume matches dot not end in other match
        if all_matches & all_followings:
            return None
        return list(all_matches), list(all_followings)

    def __and__(self, other):
        return BranchingExpression(self.expressions + [other])


class NodeExpression(Expression):
    def __init__(self, node_type: str = None, filter_fn=None, node_type_fn: Callable[[dict], str] = None):
        self.filter = filter_fn
        self.node_type = node_type
        if node_type_fn is None:
            self.node_type_fn = lambda x: x['type']
        else:
            self.node_type_fn = node_type_fn

    def _iterate_alternatives(self, nodes):
        for node in nodes:
            yield [node]

    def _match(self, nodes, graph):
        if len(nodes) != 1:
            return None

        node_name = nodes[0]
        node = graph.nodes[node_name]
        node_type = self.node_type_fn(node)
        if self.node_type == node_type:
            if self.filter and not self.filter(node):
                return None

            following = graph.successors(node_name)
            return node_name, following
        return None


def get_edge_boundaries(match: List[str], graph: nx.DiGraph):
    out_edge_boundary = list(nx.edge_boundary(graph, match, data=True))
    complement = list(filter(lambda x: x not in match, graph.nodes.keys()))
    in_edge_boundary = list(nx.edge_boundary(graph, complement, data=True))
    return in_edge_boundary, out_edge_boundary


def search_all(graph: nx.DiGraph, expression: Expression) -> List[List[str]]:
    """Returns list of node key lists that match the expression."""
    matches = []
    matched_nodes = set()
    weakly_subgraphs = [graph.subgraph(c) for c in nx.weakly_connected_components(graph)]
    for subgraph in weakly_subgraphs:
        dfs_order = nx.topological_sort(subgraph)
        for node in dfs_order:
            match, _ = expression.match([node], graph)

            if node in matched_nodes:
                continue

            if match:
                for mn in match:
                    matched_nodes.add(mn)
                matches.append(match)
    return matches



import numpy as np
import networkx as nx
from copy import deepcopy
from texttable import Texttable
from collections import deque

from nncf.quantization.layers import SymmetricQuantizer
from nncf.nncf_network import NNCFNetwork, NNCFGraph
from nncf.dynamic_graph.transform_graph import is_nncf_module
from nncf.quantization.quantizer_propagation import DEFAULT_QUANT_TRAIT_TO_OP_DICT, QuantizationTrait

class BaseMetric:
    def __init__(self):
        pass

    def collect(self):
        pass

    def get_metric_table(self):
        pass


class NetworkQuantizationShareMetric(BaseMetric):
    """
    This is a metric representing the share of the model that has been quantized.
    It includes the calculation of the following numbers:
    - Percentage of symmetric/asymmetric/per-channel/per-tensor weight quantizers relative
      to the number of placed weight quantizers
    - Percentage of symmetric/asymmetric/per-channel/per-tensor non weight quantizers relative
      to the number of placed non weight quantizers
    - Percentage of weight quantizers and non weight quantizers for each precision relative
      to the number potential* quantizers / placed quantizers
    Bitwidth distribution data is also collected.

    * The maximum possible number of potential quantizers depends on the presence of ignored
    scopes and the mode of quantizer setup that is used at the time of collecting the metric.

    """
    NAME_STR = 'NetworkQuantizationShare'

    WEIGHTS_RATIO_STR = ' WQs / All placed WQs' # WQ - weight quantizer
    ACTIVATIONS_RATIO_STR = ' AQs / All placed AQs' # AQ - activation quantizer
    TOTAL_RATIO_STR = ' Qs (out of total placed)'

    PARAMS_STR = 'Quantizer parameter'
    SYMMETRIC_STR = 'Symmetric'
    ASYMMETRIC_STR = 'Asymmetric'
    PER_CHANNEL_STR = 'Per-channel'
    SIGNED_STR = 'Signed'
    PER_TENSOR_STR = 'Per-tensor'
    UNSIGNED_STR = 'Unsigned'
    SHARE_WEIGHT_QUANTIZERS_STR = 'Placed WQs / Potential WQs'
    SHARE_ACTIVATION_QUANTIZERS_STR = 'Placed AQs / Potential AQs'

    def __init__(self, compressed_model, weights_quantizers, non_weights_quantizers, quantizer_setup_type):
        super().__init__()
        self._compressed_model = compressed_model
        self._quantizer_setup_type = quantizer_setup_type # type: QuantizerSetupType
        self.non_weights_quantizers = {k: v.quantizer_module_ref for k, v in non_weights_quantizers.items()}
        self.weights_quantizers = weights_quantizers
        self._all_quantizations = {**self.weights_quantizers, **self.non_weights_quantizers}
        self.header = [self.PARAMS_STR, self.WEIGHTS_RATIO_STR, self.ACTIVATIONS_RATIO_STR, self.TOTAL_RATIO_STR]
        self.params = {self.PER_CHANNEL_STR, self.PER_TENSOR_STR, self.UNSIGNED_STR, self.SIGNED_STR,
                       self.SYMMETRIC_STR, self.ASYMMETRIC_STR}
        self.params_bits_stat = set()
        self.num_potential_quantized_weights = len(compressed_model.get_nncf_modules())
        self.num_potential_quantized_activations = self._get_num_potential_quantized_activations()
        self.num_placed_weight_quantizers = len(self.weights_quantizers)
        self.num_placed_activation_quantizers = len(self.non_weights_quantizers)
        self.num_all_potential_quantizer = self.num_potential_quantized_weights +\
             self.num_potential_quantized_activations
        self.stat = {}
        self._ratio = {
            self.WEIGHTS_RATIO_STR: len(self.weights_quantizers),
            self.ACTIVATIONS_RATIO_STR: len(self.non_weights_quantizers),
            self.TOTAL_RATIO_STR: len(self._all_quantizations)}

    def _get_num_potential_quantized_activations(self):
        from nncf.quantization.algo import QuantizerSetupType
        retval = 0
        if self._quantizer_setup_type == QuantizerSetupType.PATTERN_BASED:
            from nncf.quantization.algo import QuantizationBuilder
            # pylint: disable=protected-access
            default_pattern = QuantizationBuilder._make_default_quantizable_subgraph_pattern()
            retval = len(self._compressed_model.get_post_pattern_insertion_points(default_pattern))
        else:
            from nncf.quantization.algo import QuantizerPropagationSolver
            insertion_point_graph = self._compressed_model.get_insertion_point_graph()
            prop_graph_solver = QuantizerPropagationSolver()
            insertion_data = prop_graph_solver.run_on_ip_graph(insertion_point_graph)
            retval = len(insertion_data)
        return retval

    def collect(self):
        for quantizer in self._all_quantizations.values():
            self.params_bits_stat.add(quantizer.num_bits)

        for h in self.header:
            self.stat[h] = {}
            for p in self.params:
                self.stat[h][p] = 0
            for p in self.params_bits_stat:
                self.stat[h][p] = 0

        for quantizer in self._all_quantizations.values():  # type: BaseQuantizer
            num_bits = quantizer.num_bits
            self.stat[self.TOTAL_RATIO_STR][num_bits] += 1
            type_ = self.WEIGHTS_RATIO_STR if quantizer.is_weights else self.ACTIVATIONS_RATIO_STR
            self.stat[type_][num_bits] += 1
            if quantizer.per_channel:
                self.stat[type_][self.PER_CHANNEL_STR] += 1
            else:
                self.stat[type_][self.PER_TENSOR_STR] += 1
            if quantizer.signed:
                self.stat[type_][self.SIGNED_STR] += 1
            else:
                self.stat[type_][self.UNSIGNED_STR] += 1
            if isinstance(quantizer, SymmetricQuantizer):
                self.stat[type_][self.SYMMETRIC_STR] += 1
            else:
                self.stat[type_][self.ASYMMETRIC_STR] += 1

    def _get_copy_statistics(self):
        statistics = deepcopy(self.stat)
        for h in self.header[1:]:
            for key, _ in statistics[h].items():
                try:
                    statistics[h][key] /= self._ratio[h]
                    statistics[h][key] *= 100
                except ZeroDivisionError:
                    statistics[h][key] = 0
        return statistics

    def get_metric_table(self):
        table_with_bits_stats = Texttable()
        table_with_other_stats = Texttable()
        data = [['Metric type', 'Value']]
        for h in (self.WEIGHTS_RATIO_STR, self.ACTIVATIONS_RATIO_STR):
            for p in self.params:
                try:
                    row = ['{} '.format(p) + str(h), '{:.2f} % ({} / {}) '.format(\
                        self.stat[h][p] / self._ratio[h] * 100, self.stat[h][p], self._ratio[h])]
                except ZeroDivisionError:
                    row = ['{} '.format(p) + h, 0]
                data.append(row)
        try:
            row = [self.SHARE_WEIGHT_QUANTIZERS_STR, '{:.2f} % ({} / {}) '.format(\
                   self.num_placed_weight_quantizers / self.num_potential_quantized_weights * 100,
                   self.num_placed_weight_quantizers, self.num_potential_quantized_weights)]
        except ZeroDivisionError:
            row = [self.SHARE_WEIGHT_QUANTIZERS_STR, '{} % '.format(0)]

        data.append(row)
        try:
            row = [self.SHARE_ACTIVATION_QUANTIZERS_STR, '{:.2f} % ({} / {}) '.format(\
                   self.num_placed_activation_quantizers / self.num_potential_quantized_activations * 100,
                   self.num_placed_activation_quantizers, self.num_potential_quantized_activations)]
        except ZeroDivisionError:
            row = [self.SHARE_ACTIVATION_QUANTIZERS_STR, '{} % '.format(0)]
        data.append(row)

        table_with_other_stats.add_rows(data)

        data = [['Num bits (N)', 'N-bits WQs / Placed WQs', 'N-bits AQs / Placed AQs', 'N-bits Qs / Placed Qs']]
        for p in self.params_bits_stat:
            row = [p]
            for h in (self.WEIGHTS_RATIO_STR, self.ACTIVATIONS_RATIO_STR, self.TOTAL_RATIO_STR):
                try:
                    row.append('{:.2f} % ({} / {}) '.format(\
                        self.stat[h][p] / self._ratio[h] * 100, self.stat[h][p], self._ratio[h]))
                except ZeroDivisionError:
                    row.append(0)
            data.append(row)
        table_with_bits_stats.add_rows(data)

        retval = {
            "Share quantization statistics:" : table_with_other_stats,
            "Bitwidth distribution:" : table_with_bits_stats
        }
        return retval

    def get_bits_stat(self):
        table = Texttable()
        data = [['Num bits (N)', 'N-bits WQs / Placed Qs', 'N-bits AQs / Placed Qs', 'N-bits Qs / Placed Qs']]
        for p in self.params_bits_stat:
            row = [p]
            for h in (self.WEIGHTS_RATIO_STR, self.ACTIVATIONS_RATIO_STR, self.TOTAL_RATIO_STR):
                try:
                    row.append(self.stat[h][p] / self._ratio[self.TOTAL_RATIO_STR] * 100)
                except ZeroDivisionError:
                    row.append(0)
            data.append(row)
        table.add_rows(data)
        return table

class MemoryCostMetric(BaseMetric):
    """

    This metric considers:
        - how many times memory consumption for network weights will decrease.
        - how many times memory consumption* for activations tensor will decrease.

    * Reflects host memory consumption, assuming only the final low-precision output activation tensors are stored
      in host memory (i.e. assuming intermediate accumulation results are only stored in device memory)

    """
    PARAMS_STR = 'params'
    NAME_STR = 'MemoryCost'

    EXPECTED_MEMORY_CONSUMPTION_DECREASE_STR = 'Memory consumption decrease for weights'
    SIZE_MEMORY_FP_WEIGHTS_STR = 'Memory consumption for full-precision weights'
    SIZE_MEMORY_COMPRESSED_WEIGHTS_STR = 'Memory consumption for quantized weights'
    MAX_MEMORY_CONSUMPTION_ACTIVATION_TENSOR_IN_FP32_MODEL_STR =\
         'Max memory consumption for an activation tensor in FP32 model'
    MAX_MEMORY_CONSUMPTION_ACTIVATION_TENSOR_IN_COMPRESSED_MODEL_STR =\
         'Max memory consumption for an activation tensor in compressed model'

    def __init__(self, compressed_model: NNCFNetwork, weights_quantizers, non_weight_quantizers):
        super().__init__()
        self._compressed_model = compressed_model
        self._weights_quantizers = weights_quantizers
        self._non_weight_quantizers = {k: v.quantizer_module_ref for k, v in non_weight_quantizers.items()}
        self.header = [self.EXPECTED_MEMORY_CONSUMPTION_DECREASE_STR, self.SIZE_MEMORY_FP_WEIGHTS_STR,\
             self.SIZE_MEMORY_COMPRESSED_WEIGHTS_STR,\
             self.MAX_MEMORY_CONSUMPTION_ACTIVATION_TENSOR_IN_FP32_MODEL_STR,\
             self.MAX_MEMORY_CONSUMPTION_ACTIVATION_TENSOR_IN_COMPRESSED_MODEL_STR]
        self.stat = {}

    def collect(self):
        self.stat[self.SIZE_MEMORY_FP_WEIGHTS_STR] = 0
        self.stat[self.SIZE_MEMORY_COMPRESSED_WEIGHTS_STR] = 0
        fp_num_bits = 32
        nncf_modules = self._compressed_model.get_nncf_modules()

        for scope_module, nncf_module in nncf_modules.items():
            count_el = np.prod(nncf_module.weight.shape)
            self.stat[self.SIZE_MEMORY_FP_WEIGHTS_STR] += count_el * fp_num_bits
            status, quantizer = self._get_quantizer_for_scope(scope_module, self._weights_quantizers)
            if status > 0:
                num_bits = quantizer.num_bits
                self.stat[self.SIZE_MEMORY_COMPRESSED_WEIGHTS_STR] += count_el * num_bits
            else:
                self.stat[self.SIZE_MEMORY_COMPRESSED_WEIGHTS_STR] += count_el * fp_num_bits
        try:
            self.stat[self.EXPECTED_MEMORY_CONSUMPTION_DECREASE_STR] = self.stat[self.SIZE_MEMORY_FP_WEIGHTS_STR] /\
             self.stat[self.SIZE_MEMORY_COMPRESSED_WEIGHTS_STR]
        except ZeroDivisionError:
            self.stat[self.EXPECTED_MEMORY_CONSUMPTION_DECREASE_STR] = 0
        self.stat[self.SIZE_MEMORY_COMPRESSED_WEIGHTS_STR] /= 2**23
        self.stat[self.SIZE_MEMORY_FP_WEIGHTS_STR] /= 2**23

        original_graph = deepcopy(self._compressed_model.get_original_graph())

        memory_consumption_fp_model = {}
        memory_consumption_compressed_model = {}
        # pylint: disable=protected-access
        original_nx_graph = original_graph._nx_graph
        nx.set_edge_attributes(original_nx_graph, 32, "precision")
        input_nodes = original_graph.get_input_nodes()
        input_node_keys = []
        for input_node in input_nodes:
            input_node_key = original_graph.get_node_key_by_id(input_node.node_id)
            input_node_keys.append(input_node_key)
            next_nodes = original_graph.get_next_nodes(input_node)
            for next_node in next_nodes:
                scope = next_node.op_exec_context.scope_in_model
                status, quantizer = self._get_quantizer_for_scope(scope, self._non_weight_quantizers)
                if status:
                    next_node_key = original_graph.get_node_key_by_id(next_node.node_id)
                    num_bits = quantizer.num_bits
                    original_nx_graph.edges[input_node_key, next_node_key]['precision'] = num_bits

        for u, v in original_nx_graph.edges:
            if u in input_node_keys:
                continue

            shape = original_nx_graph.edges[u, v][NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR]
            u_node_scope_str = str(original_nx_graph.nodes[u]['op_exec_context'].input_agnostic)
            num_bits = self.get_precision_for_activation_tensor(u, v, original_nx_graph)
            original_nx_graph.edges[u, v]['precision'] = num_bits
            memory_consumption_fp_model[u_node_scope_str] = np.prod(shape) * fp_num_bits
            memory_consumption_compressed_model[u_node_scope_str] = np.prod(shape) * num_bits
        try:
            self.stat[self.MAX_MEMORY_CONSUMPTION_ACTIVATION_TENSOR_IN_FP32_MODEL_STR] =\
                max(memory_consumption_fp_model.values()) / 2**23
            self.stat[self.MAX_MEMORY_CONSUMPTION_ACTIVATION_TENSOR_IN_COMPRESSED_MODEL_STR] =\
                max(memory_consumption_compressed_model.values()) / 2**23
        except ValueError:
            self.stat[self.MAX_MEMORY_CONSUMPTION_ACTIVATION_TENSOR_IN_FP32_MODEL_STR] = 0
            self.stat[self.MAX_MEMORY_CONSUMPTION_ACTIVATION_TENSOR_IN_COMPRESSED_MODEL_STR] = 0

    def get_precision_for_activation_tensor(self, u_node, v_node, original_nx_graph):
        scope_u_node = original_nx_graph.nodes[u_node][NNCFGraph.OP_EXEC_CONTEXT_NODE_ATTR].scope_in_model
        # pylint: disable=protected-access
        pred_u_nodes = original_nx_graph._pred[u_node]
        precision_enter_activation_tensor =\
             max([0] + [original_nx_graph.edges[pred_u_node, u_node]['precision'] for pred_u_node in pred_u_nodes])
        module = self._compressed_model.get_module_by_scope(scope_u_node)
        if is_nncf_module(module):
            status, quantizer = self._get_quantizer_for_scope(scope_u_node, self._weights_quantizers)
            if status:
                precision = max(quantizer.num_bits, precision_enter_activation_tensor)
            else:
                precision = 32
            return precision

        u_node_scope_str = str(original_nx_graph.nodes[u_node][NNCFGraph.OP_EXEC_CONTEXT_NODE_ATTR].input_agnostic)
        if u_node_scope_str in self._compressed_model.activation_quantizers:
            precision = self._compressed_model.activation_quantizers[u_node_scope_str].num_bits
        else:
            precision = precision_enter_activation_tensor
        return precision

    def _get_quantizer_for_scope(self, scope, quatizers):
        for quantizer_id, quantizer in quatizers.items():
            if quantizer_id.get_scope() == scope:
                return True, quantizer
        return False, None

    def get_metric_table(self):
        table = Texttable()
        data = [['Metric type', 'Value']]
        data.append([self.header[0], self.stat[self.header[0]]])
        for h in self.header[1:]:
            data.append([h + ' (Mbyte)', self.stat[h]])
        table.add_rows(data)

        retval = {"Memory consumption statistics:": table}
        return retval


class ShareEdgesQuantizedDataPath(BaseMetric):
    """

    This metric calculates the percentage of quantized edges relative to the total number of edges
    in the original network graph. "Quantized edge" is an edge representing a quantized activation tensor.

    """
    NAME_STR = 'ShareEdgesQuantizedDataPath'
    COUNT_QUANTIZED_EDGES_STR = 'Share edges of the quantized data path'
    QUANTIZED_EDGES_ATTR = 'quantized'
    PASSED_EDGES_ATTR = 'passed'
    NODES_GRAPH_ATTR = 'nodes'
    IS_MERGED_GRAPH_ATTR = 'is_merged'


    def __init__(self, compressed_model: NNCFNetwork):
        super().__init__()
        self._compressed_model = compressed_model
        self.stat = {}

    def collect(self):
        # pylint: disable=too-many-branches
        merged_original_graph =\
            self.get_merged_original_graph_with_patterns(self._compressed_model.get_original_graph())
        self.stat[self.COUNT_QUANTIZED_EDGES_STR] = 0
        self.header = [self.COUNT_QUANTIZED_EDGES_STR]
        nx.set_edge_attributes(merged_original_graph, False, self.QUANTIZED_EDGES_ATTR)
        nx.set_edge_attributes(merged_original_graph, False, self.PASSED_EDGES_ATTR)
        # pylint: disable=protected-access
        input_nodes = [node for node in merged_original_graph.nodes if len(merged_original_graph._pred[node]) == 0]
        queue = deque()
        for input_node in input_nodes:
            # pylint: disable=protected-access
            next_nodes = merged_original_graph._succ[input_node]
            for next_node_key in next_nodes:
                edge = merged_original_graph.edges[input_node, next_node_key]
                edge[self.PASSED_EDGES_ATTR] = True
                edge[self.QUANTIZED_EDGES_ATTR] = True
                self.stat[self.COUNT_QUANTIZED_EDGES_STR] += 1
                queue.appendleft(next_node_key)
        visited_nodes = {}
        #pylint: disable=too-many-nested-blocks
        while len(queue) != 0:
            node_key = queue.pop()
            if node_key in visited_nodes:
                continue
            if self._all_enter_edges_in_node_of_type(merged_original_graph, node_key, self.PASSED_EDGES_ATTR):
                visited_nodes[node_key] = True
                node = merged_original_graph.nodes[node_key]
                if node[self.IS_MERGED_GRAPH_ATTR]:
                    last_node = node[self.NODES_GRAPH_ATTR][-1]
                    scope_str = str(last_node[NNCFGraph.OP_EXEC_CONTEXT_NODE_ATTR].input_agnostic)
                    if scope_str in self._compressed_model.activation_quantizers:
                        self._marking_edges(merged_original_graph, node_key, queue)
                    else:
                        self._marking_edges(merged_original_graph, node_key, queue, False)
                else:
                    scope_str = str(node[NNCFGraph.OP_EXEC_CONTEXT_NODE_ATTR].input_agnostic)
                    if scope_str in self._compressed_model.activation_quantizers:
                        self._marking_edges(merged_original_graph, node_key, queue)
                    else:
                        is_op_non_change_precision_activation_tensor = True
                        node_op_name = node[NNCFGraph.OP_EXEC_CONTEXT_NODE_ATTR].operator_name
                        for op in DEFAULT_QUANT_TRAIT_TO_OP_DICT[QuantizationTrait.INPUTS_QUANTIZABLE]:
                            op_names = [op.name]
                            if op.torch_tensor_patch_spec is not None:
                                op_names = op.torch_tensor_patch_spec.underlying_function_names
                            if node_op_name in op_names:
                                is_op_non_change_precision_activation_tensor = False
                                break
                        status = is_op_non_change_precision_activation_tensor and\
                            self._all_enter_edges_in_node_of_type(merged_original_graph,\
                                node_key, self.QUANTIZED_EDGES_ATTR)
                        self._marking_edges(merged_original_graph, node_key, queue, status)
            else:
                queue.appendleft(node_key)
        self.num_merged_original_graph_edges = len(merged_original_graph.edges)

    def _get_copy_statistics(self):
        statistics = deepcopy(self.stat)
        try:
            statistics[self.COUNT_QUANTIZED_EDGES_STR] /= self.num_merged_original_graph_edges
            statistics[self.COUNT_QUANTIZED_EDGES_STR] *= 100
        except ZeroDivisionError:
            statistics[self.COUNT_QUANTIZED_EDGES_STR] = 0

        return statistics

    def _all_enter_edges_in_node_of_type(self, graph, node_key, type_edge):
        # pylint: disable=protected-access
        prev_nodes = graph._pred[node_key]
        retval = True
        for prev_node_key in prev_nodes:
            edge = graph.edges[prev_node_key, node_key]
            if not edge[type_edge]:
                retval = False
                break
        return retval

    def _marking_edges(self, graph, node_key, queue, mark=True):
        # pylint: disable=protected-access
        next_nodes = graph._succ[node_key]
        for next_node_key in next_nodes:
            edge = graph.edges[node_key, next_node_key]
            edge[self.QUANTIZED_EDGES_ATTR] = mark
            edge[self.PASSED_EDGES_ATTR] = True
            queue.appendleft(next_node_key)
            if mark:
                self.stat[self.COUNT_QUANTIZED_EDGES_STR] += 1

    def get_metric_table(self):
        table = Texttable()
        data = [['Metric type', 'Value']]
        try:
            data.append([self.header[0], '{:.2f} % ({} / {})'.format(
                self.stat[self.COUNT_QUANTIZED_EDGES_STR] / self.num_merged_original_graph_edges * 100,
                self.stat[self.COUNT_QUANTIZED_EDGES_STR], self.num_merged_original_graph_edges)])
        except ZeroDivisionError:
            data.append([self.header[0], '{} % '.format(0)])
        table.add_rows(data)

        retval = {"Quantization configuration statistics:" : table}
        return retval

    def get_merged_original_graph_with_patterns(self, original_graph: NNCFGraph):
        import nncf.dynamic_graph.patterns as p
        from nncf.dynamic_graph.graph_matching import search_all

        pattern = p.LINEAR_OPS + p.ANY_BN_ACT_COMBO | p.LINEAR_OPS + p.ELTWISE_UNIFORM_OPS
        # pylint: disable=protected-access
        matches = search_all(original_graph._nx_graph, pattern)
        merged_graph = deepcopy(original_graph._nx_graph)
        nx.set_node_attributes(merged_graph, False, self.IS_MERGED_GRAPH_ATTR)
        for match in matches:
            if len(match) == 1:
                continue

            input_node_key = match[0]
            output_node_key = match[-1]
            in_edges = list(merged_graph.in_edges(input_node_key))
            out_edges = list(merged_graph.out_edges(output_node_key))

            in_edge_copies_dict = {}
            for in_edge_key in in_edges:
                in_edge_copies_dict[in_edge_key] = deepcopy(merged_graph.edges[in_edge_key])
            out_edge_copies_dict = {}
            for out_edge_key in out_edges:
                out_edge_copies_dict[out_edge_key] = deepcopy(merged_graph.edges[out_edge_key])

            merged_node_key = ""
            merged_nodes = []
            for node_key in match:
                merged_node_key += node_key + '\n'
                # pylint: disable=protected-access
                merged_nodes.append(original_graph._nx_graph.nodes[node_key])
                merged_graph.remove_node(node_key)
            merged_node_attrs = {
                NNCFGraph.KEY_NODE_ATTR: merged_node_key,
                self.NODES_GRAPH_ATTR: merged_nodes,
                self.IS_MERGED_GRAPH_ATTR: True
            }
            merged_graph.add_node(merged_node_key, **merged_node_attrs)
            for in_edge_key, in_edge_attrs in in_edge_copies_dict.items():
                merged_graph.add_edge(in_edge_key[0], merged_node_key, **in_edge_attrs)
            for out_edge_key, out_edge_attrs in out_edge_copies_dict.items():
                merged_graph.add_edge(merged_node_key, out_edge_key[1], **out_edge_attrs)

        return merged_graph

    @staticmethod
    def visualize_marked_graph(merged_original_graph):
        out_graph = nx.DiGraph()
        for node_key, _ in merged_original_graph.nodes.items():
            out_graph.add_node(node_key)
        for u, v in merged_original_graph.edges:
            edge = merged_original_graph.edges[u, v]
            if edge[ShareEdgesQuantizedDataPath.QUANTIZED_EDGES_ATTR]:
                attrs = {"color": "blue"}
            out_graph.add_edge(u, v, **attrs)
        return out_graph



from collections import OrderedDict
from inspect import Signature, Parameter
from typing import List

import torch

from nncf.dynamic_graph.patch_pytorch import register_operator
from nncf.dynamic_graph.graph_builder import ModelInputInfo, create_mock_tensor
from nncf.utils import is_tensor, objwalk
from nncf.nncf_logger import logger as nncf_logger


@register_operator()
def nncf_model_input(tensor: 'torch.Tensor'):
    return tensor


def wrap_nncf_model_inputs_with_objwalk(model_args, model_kwargs):
    model_args = objwalk(model_args, is_tensor, nncf_model_input)
    model_kwargs = objwalk(model_kwargs, is_tensor, nncf_model_input)
    return model_args, model_kwargs


class InputInfoWrapManager:
    INPUTS_MISMATCH_WARNING_TEXT = "Compression with regards to this input may occur incorrectly. Make sure " \
                                   "you call the compressed model with inputs that correspond to what NNCF was " \
                                   "configured to expect (either via NNCF config's input_infos, or custom" \
                                   "dummy_forward_fn/wrap_inputs_fn parameters), or that you know what you are " \
                                   "doing. This warning will not be shown again."
    ARGS_INPUTS_MISMATCH_FORMAT_STRING = "Inputs mismatch - could not find arg with idx {} in NNCF-wrapped model " \
                                         "input args! " + INPUTS_MISMATCH_WARNING_TEXT
    KWARGS_INPUTS_MISMATCH_FORMAT_STRING = "Inputs mismatch - could not find kwarg '{}' in NNCF-wrapped model input " \
                                           "kwargs! " + INPUTS_MISMATCH_WARNING_TEXT

    def __init__(self, input_infos: List[ModelInputInfo],
                 fwd_signature: Signature,
                 module_ref_for_device: torch.nn.Module = None):
        self._module_ref_for_device = module_ref_for_device
        arg_iis_list = [ii for ii in input_infos if ii.keyword is None]
        kwarg_iis_list = [(ii.keyword, ii) for ii in input_infos if ii.keyword is not None]
        kwarg_iis = OrderedDict()
        arg_iis = tuple(arg_iis_list)
        for kw, ii in kwarg_iis_list:
            kwarg_iis[kw] = ii
        bound_params = fwd_signature.bind(*arg_iis, **kwarg_iis)

        self._fwd_params_to_input_infos_odict = bound_params.arguments
        self._fwd_signature = fwd_signature  # type: Signature

    def set_device(self, device: str):
        self._device = device

    def wrap_inputs(self, model_args, model_kwargs):
        bound_model_params = self._fwd_signature.bind(*model_args, **model_kwargs)
        for param_name in self._fwd_params_to_input_infos_odict:
            param_kind = self._fwd_signature.parameters[param_name].kind
            if param_kind is Parameter.VAR_POSITIONAL or param_kind is Parameter.VAR_KEYWORD:
                nncf_logger.warning("An input_info tensor was bound to a *args or **kwargs variadic parameter in the"
                                    "forward's signature! This is currently unsupported by NNCF. Input compression may "
                                    "be incorrect.")
                # Currently won't support input info mapping to *args or **kwargs-mapped parameters
                continue

            if param_name not in bound_model_params.arguments:
                nncf_logger.warning("A call to a compressed model's forward occured without one of the params"
                                    "specified in input_infos! Input compression may be incorrect. Trying to recover "
                                    "by wrapping the default value for the parameter.")
                bound_model_params.apply_defaults()

            potential_tensor = bound_model_params.arguments[param_name]
            if potential_tensor is not None:
                bound_model_params.arguments[param_name] = nncf_model_input(bound_model_params.arguments[param_name])
            else:
                # Default was None - cannot wrap as-is. Will wrap a dummy tensor as specified in
                # input infos - will conserve the call order of nncf_model_input nodes,
                # and the post-hooks for the input node will execute. The result won't go anywhere, though.
                nncf_logger.warning("Wrapping a dummy tensor for input {}".format('param_name'))
                info_for_missing_input = self._fwd_params_to_input_infos_odict[param_name]
                device = 'cuda'
                if self._module_ref_for_device is not None:
                    device = next(self._module_ref_for_device.parameters()).device
                dummy_tensor = create_mock_tensor(info_for_missing_input, device)
                _ = nncf_model_input(dummy_tensor)

        return bound_model_params.args, bound_model_params.kwargs


# Access via _original op because by this moment the nncf_model_input name will already be wrapped by wrap_operator
# and its __name__ attribute changed correspondingly.
# pylint:disable=protected-access
MODEL_INPUT_OP_NAME = nncf_model_input._original_op.__name__



"""
 Copyright (c) 2020 Intel Corporation
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
      http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
from typing import List

import torch
from texttable import Texttable

from nncf.algo_selector import COMPRESSION_ALGORITHMS
from nncf.compression_method_api import CompressionAlgorithmController, CompressionLevel
from nncf.layers import NNCF_CONV_MODULES_DICT
from nncf.nncf_logger import logger as nncf_logger
from nncf.nncf_network import NNCFNetwork
from nncf.pruning.base_algo import BasePruningAlgoBuilder, PrunedModuleInfo, BasePruningAlgoController
from nncf.pruning.export_helpers import ModelPruner
from nncf.pruning.filter_pruning.functions import calculate_binary_mask, FILTER_IMPORTANCE_FUNCTIONS, \
    tensor_l2_normalizer
from nncf.pruning.filter_pruning.layers import FilterPruningBlock, inplace_apply_filter_binary_mask
from nncf.pruning.schedulers import PRUNING_SCHEDULERS
from nncf.pruning.utils import get_rounded_pruned_element_number


@COMPRESSION_ALGORITHMS.register('filter_pruning')
class FilterPruningBuilder(BasePruningAlgoBuilder):
    def create_weight_pruning_operation(self, module):
        return FilterPruningBlock(module.weight.size(0))

    def build_controller(self, target_model: NNCFNetwork) -> CompressionAlgorithmController:
        return FilterPruningController(target_model,
                                       self._pruned_module_info,
                                       self.config)

    def _is_pruned_module(self, module):
        # Currently prune only Convolutions
        return isinstance(module, tuple(NNCF_CONV_MODULES_DICT.keys()))

    def get_types_of_pruned_modules(self):
        types = [str.lower(v.__name__) for v in NNCF_CONV_MODULES_DICT.values()]
        return types


class FilterPruningController(BasePruningAlgoController):
    def __init__(self, target_model: NNCFNetwork,
                 pruned_module_info: List[PrunedModuleInfo],
                 config):
        super().__init__(target_model, pruned_module_info, config)
        params = self.config.get("params", {})
        self.frozen = False
        self.pruning_rate = 0

        self.weights_normalizer = tensor_l2_normalizer  # for all weights in common case
        self.filter_importance = FILTER_IMPORTANCE_FUNCTIONS.get(params.get('weight_importance', 'L2'))
        self.all_weights = params.get("all_weights", False)
        scheduler_cls = PRUNING_SCHEDULERS.get(params.get("schedule", "baseline"))
        self._scheduler = scheduler_cls(self, params)

    @staticmethod
    def _get_mask(minfo: PrunedModuleInfo):
        return minfo.operand.binary_filter_pruning_mask

    def statistics(self):
        stats = super().statistics()
        stats['pruning_rate'] = self.pruning_rate
        return stats

    def freeze(self):
        self.frozen = True

    def set_pruning_rate(self, pruning_rate):
        self.pruning_rate = pruning_rate
        if not self.frozen:
            if self.all_weights:
                self._set_binary_masks_for_all_pruned_modules()
            else:
                self._set_binary_masks_for_filters()
            if self.zero_grad:
                self.zero_grads_for_pruned_modules()
        self._apply_masks()
        self.run_batchnorm_adaptation(self.config)

    def _set_binary_masks_for_filters(self):
        nncf_logger.debug("Setting new binary masks for pruned modules.")

        with torch.no_grad():
            for minfo in self.pruned_module_info:
                pruning_module = minfo.operand
                # 1. Calculate importance for all filters in all weights
                # 2. Calculate thresholds for every weight
                # 3. Set binary masks for filter
                filters_importance = self.filter_importance(minfo.module.weight)
                num_of_sparse_elems = get_rounded_pruned_element_number(filters_importance.size(0),
                                                                        self.pruning_rate)
                threshold = sorted(filters_importance)[num_of_sparse_elems]
                pruning_module.binary_filter_pruning_mask = calculate_binary_mask(filters_importance, threshold)

    def _set_binary_masks_for_all_pruned_modules(self):
        nncf_logger.debug("Setting new binary masks for all pruned modules together.")

        normalized_weights = []
        filter_importances = []
        for minfo in self.pruned_module_info:
            pruning_module = minfo.operand
            # 1. Calculate importance for all filters in all weights
            # 2. Calculate thresholds for every weight
            # 3. Set binary masks for filter
            normalized_weight = self.weights_normalizer(minfo.module.weight)
            normalized_weights.append(normalized_weight)

            filter_importances.append(self.filter_importance(normalized_weight))
        importances = torch.cat(filter_importances)
        threshold = sorted(importances)[int(self.pruning_rate * importances.size(0))]

        for i, minfo in enumerate(self.pruned_module_info):
            pruning_module = minfo.operand
            pruning_module.binary_filter_pruning_mask = calculate_binary_mask(filter_importances[i], threshold)

    def _apply_masks(self):
        nncf_logger.debug("Applying pruning binary masks")

        def _apply_binary_mask_to_module_weight_and_bias(module, mask, module_name=""):
            with torch.no_grad():
                # Applying mask to weights
                inplace_apply_filter_binary_mask(mask, module.weight, module_name)
                # Applying mask to bias too (if exists)
                if module.bias is not None:
                    inplace_apply_filter_binary_mask(mask, module.bias, module_name)

        for minfo in self.pruned_module_info:
            _apply_binary_mask_to_module_weight_and_bias(minfo.module, minfo.operand.binary_filter_pruning_mask,
                                                         minfo.module_name)

            # Applying mask to the BatchNorm node
            related_modules = minfo.related_modules
            if minfo.related_modules is not None and PrunedModuleInfo.BN_MODULE_NAME in minfo.related_modules \
                    and related_modules[PrunedModuleInfo.BN_MODULE_NAME] is not None:
                bn_module = related_modules[PrunedModuleInfo.BN_MODULE_NAME]
                _apply_binary_mask_to_module_weight_and_bias(bn_module, minfo.operand.binary_filter_pruning_mask)

            if minfo.related_modules is not None and PrunedModuleInfo.DEPTHWISE_BN_NAME in minfo.related_modules \
                    and related_modules[PrunedModuleInfo.DEPTHWISE_BN_NAME] is not None:
                bn_module = related_modules[PrunedModuleInfo.DEPTHWISE_BN_NAME]
                _apply_binary_mask_to_module_weight_and_bias(bn_module, minfo.operand.binary_filter_pruning_mask)

    @staticmethod
    def create_stats_table_for_pruning_export(old_modules_info, new_modules_info):
        """
        Creating a table with comparison of model params number before and after pruning.
        :param old_modules_info: Information about pruned layers before actually prune layers.
        :param new_modules_info: Information about pruned layers after actually prune layers.
        """
        table = Texttable()
        header = ["Name", "Weight's shape", "New weight's shape", "Bias shape", "New bias shape",
                  "Weight's params count", "New weight's params count",
                  "Mask zero %", "Layer PR"]
        data = [header]

        for layer in old_modules_info.keys():
            assert layer in new_modules_info

            drow = {h: 0 for h in header}
            drow["Name"] = layer
            drow["Weight's shape"] = old_modules_info[layer]['w_shape']
            drow["New weight's shape"] = new_modules_info[layer]['w_shape']
            drow["Bias shape"] = old_modules_info[layer]['b_shape']
            drow["New bias shape"] = new_modules_info[layer]['b_shape']

            drow["Weight's params count"] = old_modules_info[layer]['params_count']
            drow["New weight's params count"] = new_modules_info[layer]['params_count']

            drow["Mask zero %"] = old_modules_info[layer]['mask_pr']

            drow["Layer PR"] = 1 - new_modules_info[layer]['params_count'] / old_modules_info[layer]['params_count']
            row = [drow[h] for h in header]
            data.append(row)
        table.add_rows(data)
        return table

    def prepare_for_export(self):
        """
        This function discards the pruned filters based on the binary masks
        before exporting the model to ONNX.
        """
        self._apply_masks()
        model = self._model.eval().cpu()
        graph = model.get_original_graph()
        # pylint: disable=protected-access
        nx_graph = graph._nx_graph

        parameters_count_before = model.get_parameters_count_in_model()
        flops = model.get_MACs_in_model()
        pruned_layers_stats = self.get_stats_for_pruned_modules()

        model_pruner = ModelPruner(model, graph, nx_graph)
        model_pruner.prune_model()

        parameters_count_after = model.get_parameters_count_in_model()
        flops_after = model.get_MACs_in_model()
        new_pruned_layers_stats = self.get_stats_for_pruned_modules()
        stats = self.create_stats_table_for_pruning_export(pruned_layers_stats, new_pruned_layers_stats)

        nncf_logger.info(stats.draw())
        nncf_logger.info('Final Model Pruning Rate = %.3f', 1 - parameters_count_after / parameters_count_before)
        nncf_logger.info('Total MAC pruning level = %.3f', 1 - flops_after / flops)

    def compression_level(self) -> CompressionLevel:
        target_pruning_level = self.scheduler.pruning_target
        actual_pruning_level = self.pruning_rate
        if actual_pruning_level == 0:
            return CompressionLevel.NONE
        if actual_pruning_level >= target_pruning_level:
            return CompressionLevel.FULL
        return CompressionLevel.PARTIAL
